import re
import logging
import calendar
import datetime
from marshmallow import Schema, fields, EXCLUDE, INCLUDE
from sqlalchemy import (Column, Integer, Unicode, DateTime, ForeignKey,
    Boolean, func)
from sqlalchemy.orm import relationship
from potnanny.utils import utcnow
from potnanny.database import Base
from potnanny.models.mixins import CRUDMixin
from potnanny.models.ext import MutableDict, JSONEncodedDict
from potnanny.models.weekday import WeekdayMap
from potnanny.controllers.outlet import switch_device_outlet

logger = logging.getLogger(__name__)


class ScheduleSchema(Schema):
    class Meta:
        unknown = EXCLUDE

    name = fields.String(allow_none=False)
    device_id = fields.Integer(allow_none=True)
    outlet = fields.Integer(allow_none=False)
    on_time = fields.String(allow_none=False)
    off_time = fields.String(allow_none=False)
    days = fields.Integer(allow_none=False)


class Schedule(Base, CRUDMixin):
    """
    Time is like: "13:30" (24h/UTC based)
    Days is an INT, based on:
     sunday=64, mon=32, tue=16, wed=8, thu=4, fri=2, sat=1
    """

    __tablename__ = 'schedules'

    id = Column(Integer, primary_key=True)
    name = fields.String(nullable=False)
    outlet = Column(Integer, nullable=False, default=1)
    days = Column(Integer, nullable=False, default=127)
    on_time = Column(Unicode(16), nullable=False)
    off_time = Column(Unicode(16), nullable=False)
    created = Column(DateTime, server_default=func.now())

    # relationships
    device_id = Column(Integer, ForeignKey('devices.id'))


    def __repr__(self):
        return "<Schedule ({}, {}, {})>".format(self.id, self.on_time, self.off_time)


    def as_dict(self):
        return {
            'id': self.id,
            'days': self.days,
            'on_time': self.on_time,
            'off_time': self.off_time,
            'outlet': self.outlet,
            'device_id': self.device_id,
            'created': self.created.isoformat() + "Z",
        }


    def runs_now(self, now=None, show_key=False):
        """
        Does this schedule need to run now? True|False
        """

        if now is None or type(now) is not datetime.datetime:
            now = datetime.datetime.now()

        dow = calendar.day_name[now.weekday()]
        if dow not in WeekdayMap.weekdays_from_value(self.days):
            return False

        for key in ['on', 'off']:
            label = "%s_time" % key
            if not hasattr(self, label):
                continue

            try:
                hh, mm = getattr(self, label).split(':')
                if int(hh) == now.hour and int(mm) == now.minute:
                    if show_key is True:
                        return (True, key)
                    else:
                        return True
            except Exception as x:
                logger.warning(x)

        if show_key is True:
            return (False, None)
        else:
            return False


    async def run(self, now=None):
        """
        Run this schedule now
        """

        if now is None or type(now) is not datetime.datetime:
            # schedule times are stored according to user local time,
            # so we cannot use UTC here. This... shoud be normalized someday,
            # but its a pain in the ass otherwise.
            now = datetime.datetime.now()

        true, key = self.runs_now(now, True)
        if not true:
            return

        try:
            state = 1
            if (key == 'off'):
                state = 0

            logger.debug("Switching device %d, outlet %d %s" % (
                self.device_id, self.outlet, key.upper()))
            rval = await switch_device_outlet(
                self.device_id, self.outlet, state)
        except Exception as x:
            logger.warning(x)

        return



from .base import (DevicePlugin, ActionPlugin, PipelinePlugin,
    BluetoothDevicePlugin, GPIODevicePlugin)
from .mixins import PluginMixin, FingerprintMixin, KeychainMixin

__all__ = [DevicePlugin, BluetoothDevicePlugin, GPIODevicePlugin, ActionPlugin,
    PipelinePlugin, PluginMixin, FingerprintMixin, KeychainMixin]



import logging
from marshmallow import Schema, fields, EXCLUDE, INCLUDE
from sqlalchemy import (Column, Integer, DateTime, Unicode, Boolean,
    UnicodeText, ForeignKey, func)
from sqlalchemy.orm import relationship
from potnanny.database import Base
from potnanny.plugins.mixins import PluginMixin
from potnanny.models.mixins import CRUDMixin
from potnanny.models.ext import MutableDict, JSONEncodedDict

logger = logging.getLogger(__name__)


class DeviceSchema(Schema):
    class Meta:
        unknown = EXCLUDE

    name = fields.String()
    notes = fields.String(allow_none=True)
    room_id = fields.Integer(allow_none=True)
    interface = fields.String()
    attributes = fields.Dict(allow_none=True)


class Device(Base, CRUDMixin, PluginMixin):
    __tablename__ = 'devices'

    id = Column(Integer, primary_key=True)
    name = Column(Unicode(64), nullable=False, unique=False)
    interface = Column(Unicode(128), nullable=True, unique=False)
    attributes = Column(MutableDict.as_mutable(JSONEncodedDict), nullable=True)
    created = Column(DateTime, server_default=func.now())
    modified = Column(DateTime, server_default=func.now())

    # make relationships compatible with asyncio sessions
    __mapper_args__ = {"eager_defaults": True}

    # relationships
    room_id = Column(Integer, ForeignKey('rooms.id'), nullable=True)
    measurements = relationship('Measurement',
        backref='device', cascade='all,delete')
    schedules = relationship('Schedule',
        backref='device', cascade='all,delete')
    controls = relationship('Control',
        backref='device', cascade='all,delete')
    actionss = relationship('Action',
        backref='device', cascade='all,delete')

    def __repr__(self):
        return "<Device ({})>".format(self.name)

    def as_dict(self):
        data = {
            'id': self.id,
            'name': self.name,
            'created': self.created.isoformat() + "Z",
            'room_id': self.room_id,
            'interface': self.interface,
            'attributes': self.attributes,
        }

        try:
            data['controls'] = self.controls
        except:
            pass

        try:
            data['schedules'] = self.schedules
        except:
            pass

        try:
            mylist = self.plugin.reports
            if ('temperature' in mylist and
                'humidity' in mylist and 'vpd' not in mylist):
                mylist.append('vpd')
            data['reports'] = mylist
        except:
            pass

        try:
            for k, v in self.plugin.__dict__.items():
                if 'plugin' not in data:
                    data['plugin'] = {}

                if k.startswith('_'):
                    continue

                data['plugin'][k] = v
        except Exception as x:
            logger.debug(x)

        return data


    async def poll(self):
        """
        Poll our plugin instance for measurements, if possible
        args:
            none
        returns:
            dict, or None
        """

        if (not hasattr(self.plugin, 'is_pollable') or
            self.plugin.is_pollable is not True):
            return None

        results = {
            'id': self.id,
            'name': self.name,
            'values': {},
        }

        try:
            logger.debug("polling plugin instance: %s" % self.plugin)
            values = await self.plugin.poll()
            results['values'] = values
        except Exception as x:
            logger.warning(str(x))

        return results


    def read_advertisement(self, device, advertisement):
        """
        Use plugin instance to decode the BLE advertisement data
        args:
            the device reporting data
            the advertisement data
        returns:
            dict, or None
        """

        if (not hasattr(self.plugin, 'is_reader') or
            self.plugin.is_reader is not True):
            return None

        results = {
            'id': self.id,
            'name': self.name,
            'values': {},
        }

        try:
            values = self.plugin.read_advertisement(device, advertisement)
            if values:
                results['values'] = values
        except:
            pass

        return results


    async def on(self, outlet=1):
        """
        Switch device ON
        """

        if (not hasattr(self.plugin, 'is_switchable') or
            self.plugin.is_switchable is not True):
            return None

        try:
            result = await self.plugin.on(outlet)
            return result
        except Exception as x:
            logger.warning(str(x))
            return None


    async def off(self, outlet=1):
        """
        Switch device OFF
        """

        if (not hasattr(self.plugin, 'is_switchable') or
            self.plugin.is_switchable is not True):
            return None

        try:
            result = await self.plugin.off(outlet)
            return result
        except Exception as x:
            logger.warning(str(x))
            return None


    async def set_value(self, data):
        """
        *FUTURE*
        Set key/value data on device.
        args:
            - dict
        returns:
        """

        if (not hasattr(self.plugin, 'is_settable') or
            self.plugin.is_settable is not True):
            return None

        if type(data) is not dict:
            logger.warning("Expecting dict but got %s" % type(data))
            return None

        try:
            result = await self.plugin.set_value(data)
            return result
        except Exception as x:
            logger.warning(x)
            return None



import asyncio
import datetime
import logging
import random
from sqlalchemy import func
from sqlalchemy.sql.expression import select
from potnanny.models.measurement import Measurement
from potnanny.models.device import Device
from potnanny.models.interface import ObjectInterface, execute_statement
from potnanny.controllers.poll import Poller
from potnanny.locks import LOCKS


logger = logging.getLogger(__name__)


async def switch_device_outlet(device_id: int, outlet_id: int, state: int):
    """
    Convenience function for times when we need a basic function
    like... a functools.partial call
    """

    ctl = OutletControl(device_id, outlet_id)
    result = await ctl.switch(state)
    return result


class OutletControl:
    """
    Class to interface with device switchable outlets
    """

    def __init__(self, device_id: int, outlet_id: int):
        self.device_id = device_id
        self.outlet_id = outlet_id
        self.lock = LOCKS['bluetooth']
        self.device = None


    async def switch(self, state: int, force: bool = False) -> int:
        """
        Switch device to requested state
        """

        if not force:
            latest = await self.last_state()
            if latest == state:
                return state

        success = False
        tries = 3
        while success is not True and tries:
            try:
                if self.lock:
                    async with self.lock:
                        success = await self._switch_outlet(state)
                else:
                    success = await self._switch_outlet(state)

                if success:
                    break
            except:
                pass

            tries -= 1
            await asyncio.sleep(float("%0.1f" % random.random()))

        if not success:
            logger.warning(
                f"Switch fail device {self.device_id}, outlet {self.outlet_id}"
            )
            return -1
        else:
            # use the poller to record the new state of the outlet
            now = datetime.datetime.utcnow().replace(microsecond=0)
            p = Poller(now=now)
            t1 = asyncio.create_task(p.poll_id(self.device_id))
            # result = await self._record_state(state)
            # if not result:
            #    logger.warning("Failed to record outlet switched state")

            return state


    async def last_state(self, threshold: int = 30) -> int:
        """
        Get most recent recorded db measurement of outlet state
        return will be the last state (0|1), or -1 if could not be found.
        """

        now = datetime.datetime.utcnow().replace(microsecond=0)
        start = now - datetime.timedelta(seconds=threshold)

        stmt = select(
            Measurement.value, func.max(Measurement.created)
        ).filter(
            Measurement.device_id == self.device_id
        ).filter(
            Measurement.type == f"outlet_{self.outlet_id}"
        ).group_by(
            Measurement.device_id
        )

        try:
            results = await execute_statement(stmt)
            row = results.first()
            # exclude the record if it older than threshold seconds
            if row[1] < start:
                return -1
            return int(row[0])
        except:
            return -1


    async def _record_state(self, state: int) -> bool:
        """
        Record the new device outlet state in the measurement table
        """

        now = datetime.datetime.utcnow().replace(microsecond=0)
        opts = {
            'device_id': self.device_id,
            'type': f"outlet_{self.outlet_id}",
            'value': state,
            'created': now }

        try:
            meas = Measurement(**opts)
            await meas.insert()
            return True
        except:
            return False


    async def _switch_outlet(self, state):
        if self.device is None:
            try:
                self.device = await ObjectInterface(Device).get_by_id(
                    self.device_id)
            except:
                logger.warning(f"No suitable device found (id {self.device_id})")
                return False

        if not hasattr(self.device.plugin, 'is_switchable') or \
            self.device.plugin.is_switchable is not True:
            logger.warning(f"Device with id {device_id} is not switchable")
            return False

        if bool(state):
            rval = await self.device.on(self.outlet_id)
            if int(rval) == 1:
                return True
        else:
            rval = await self.device.off(self.outlet_id)
            if int(rval) == 0:
                return True

        return False



import asyncio
import copy
import logging
from potnanny.plugins import PipelinePlugin

logger = logging.getLogger(__name__)


class Pipeline:
    """
    All new measurements go into the pipline and delivered to targets
    """

    def __init__(self, *args, **kwargs):
        for k,v in kwargs.items():
            if hasattr(self, k):
                setattr(self, k, v)


    async def input(self, measurements):
        """
        Input measurements into the pipeline for processing
        """

        tasks = []
        for p in PipelinePlugin.plugins:
            logger.debug("Queueing pipeline job: %s" % p)
            tasks.append(p().input(copy.deepcopy(measurements)))

        if not tasks:
            return

        try:
            await asyncio.gather(*tasks)
        except Exception as x:
            logger.warning(x)



import os
import unittest
from potnanny.utils.paths import resolve_path

class TestUtilFuncs(unittest.TestCase):
# ###############################################

    def test_path(self):
        result = resolve_path('~')
        assert result == os.path.abspath(os.getenv('HOME'))


if __name__ == '__main__':
    unittest.main()



import asyncio
import aiosmtplib
import unittest
import datetime
import potnanny.database as db
from email.mime.text import MIMEText
from unittest import IsolatedAsyncioTestCase
from potnanny.models.room import Room
from potnanny.models.device import Device
from potnanny.models.measurement import Measurement
from potnanny.models.action import Action
from potnanny.models.error import Error
from potnanny.models.trigger import Trigger
from potnanny.models.keychain import Keychain
from potnanny.models.interface import ObjectInterface
from potnanny.plugins import ActionPlugin
from potnanny.controllers.pipeline import Pipeline
from potnanny.plugins import PipelinePlugin



class ActionPipeline(PipelinePlugin):
    """
    Class to route pipeline measurements to controls
    """

    name = "Action Pipeline Plugin"
    description = "Route measurements to Actions"

    def __init__(self, *args, **kwargs):
        pass


    async def input(self, measurements):
        tasks = []
        objects = await ObjectInterface(Action).get_all()
        for obj in objects:
            for m in measurements:

                # pre-filter, only appropriate measurements go to the action
                try:
                    if obj.device_id != m['device_id']:
                        continue

                    if obj.attributes['type'] != m['type']:
                        continue

                    tasks.append(obj.input(m))
                except Exception as x:
                    print(x)

        if not len(tasks):
            return

        try:
            await asyncio.gather(*tasks)
        except Exception as x:
            print(x)




class TestActionModel(IsolatedAsyncioTestCase):
# ###############################################

    async def asyncSetUp(self):
        uri = 'sqlite+aiosqlite://'
        await db.init_db(uri)
        try:
            room = Room(name="test room")
            await room.insert()

            device = Device(
                name="test device",
                interface="Foo",
                room_id=1)
            await device.insert()

            action1 = Action(
                name="test action 1",
                interface="__main__.GmailSender",
                device_id=1,
                attributes={
                    "type": "temperature",
                    "condition": "gt",
                    "threshold": 80
                }
            )
            await action1.insert()

            action2 = Action(
                name="test action 2",
                interface="__main__.GmailSender",
                device_id=1,
                attributes={
                    "type": "humidity",
                    "condition": "gt",
                    "threshold": 50,
                    "sleep_minutes": 60
                }
            )
            await action2.insert()

            now = datetime.datetime.utcnow()
            then = now - datetime.timedelta(minutes=30)
            trigger = Trigger(action_id=2, created=then)
            await trigger.insert()

        except:
            pass


    async def asyncTearDown(self):
        pass

    async def test_create(self):
        measurements = [
            {"type": "humidity", "value": 55.0, "device_id": 1},
            {"type": "battery", "value": 100.0, "device_id": 1},
            {"type": "vpd", "value": 1.0, "device_id": 1},
            {"type": "temperature", "value": 75.0, "device_id": 1},
            {"type": "temperature", "value": 85.0, "device_id": 1},
        ]
        p = Pipeline()
        await p.input(measurements)

if __name__ == '__main__':
    unittest.main()



import asyncio
import logging
from potnanny.models.device import Device
from potnanny.models.interface import ObjectInterface
from potnanny.controllers.pipeline import Pipeline
from potnanny.utils import utcnow
from .ble import BLEInterface
from .gpio import GPIOInterface
from .parse import Parser


logger = logging.getLogger(__name__)


class Poller:

    def __init__(self, *args, **kwargs):
        self.convert_c = False
        self.leaf_offset = -2
        self.devices = []
        self.now = utcnow()
        self._kwargs = kwargs

        for k,v in kwargs.items():
            if hasattr(self, k):
                setattr(self, k, v)


    async def poll(self):
        measurements = []

        if not self.devices:
            await self._load_devices()

        tasks = [
            BLEInterface().poll_devices(self.devices),
            GPIOInterface().poll_devices(self.devices)
        ]

        # collect measurements
        results = await asyncio.gather(*tasks)

        # parse and normalize measurement data
        if results:
            measurements = Parser(**self._kwargs).parse(results)
            logger.debug("MEASUREMENTS: %s" % measurements)

        # send measurement data into the pipeline
        p = Pipeline()
        await p.input(measurements)
        return measurements


    async def poll_id(self, pk:int):
        """
        Poll measurements only for a named device id
        """

        device = await ObjectInterface(Device).get_by_id(pk)
        logger.debug("loaded individual device: {device}")
        if device is not None:
            self.devices.append(device)
            await self.poll()


    async def _load_devices(self):
        devices = await ObjectInterface(Device).get_all()

        # skip devices not assigned to rooms yet
        for d in devices:
            if d.room_id is not None:
                self.devices.append(d)



import os
import asyncio
import logging
import jinja2
import aiohttp_jinja2
import aiohttp_cors
import base64
from cryptography.fernet import Fernet
from aiohttp import web
from aiohttp_session import setup, SimpleCookieStorage
from aiohttp_session.cookie_storage import EncryptedCookieStorage
from .index import routes as index_routes
from .auth import routes as auth_routes
from .users import routes as user_routes
from .keychains import routes as keychain_routes
from .rooms import routes as room_routes
from .devices import routes as device_routes
from .environments import routes as environment_routes
from .settings import routes as setting_routes
from .actions import routes as action_routes
from .controls import routes as control_routes
from .schedules import routes as schedule_routes
from .utils import routes as util_routes
from .plugins import routes as plugin_routes
from .tests import routes as test_routes
from .wifi import routes as wifi_routes


logger = logging.getLogger(__name__)


def init_api():
    logger.debug("Initializing Web API")

    app = web.Application()

    # setup static path handling
    app['static_root_url'] = '/static'
    static_path = os.path.join(os.path.dirname(__file__), "static")
    app.router.add_static('/static', static_path, name='static')

    # session cookie storage
    key = Fernet.generate_key()
    secret = base64.urlsafe_b64decode(key)
    setup(app, EncryptedCookieStorage(secret,
        cookie_name='POTNANNY_API',
        samesite="None",
        secure=True
        )
    )

    # simple cookie storage for devel ONLY
    # setup(app, SimpleCookieStorage())

    # plug in jinja template handling
    aiohttp_jinja2.setup(app,
        loader=jinja2.FileSystemLoader(
            os.path.join(os.path.dirname(__file__), "templates")))

    # add routing to endpoints
    app.add_routes(index_routes)
    app.add_routes(auth_routes)
    app.add_routes(user_routes)
    app.add_routes(room_routes)
    app.add_routes(device_routes)
    app.add_routes(keychain_routes)
    app.add_routes(environment_routes)
    app.add_routes(setting_routes)
    app.add_routes(util_routes)
    app.add_routes(plugin_routes)
    app.add_routes(action_routes)
    app.add_routes(control_routes)
    app.add_routes(schedule_routes)
    app.add_routes(test_routes)
    app.add_routes(wifi_routes)


    logger.debug("Building CORS routes")
    cors = aiohttp_cors.setup(app, defaults={
        "*": aiohttp_cors.ResourceOptions(
            allow_credentials=True,
            expose_headers="*",
            allow_headers="*",
        )
    })
    for route in list(app.router.routes()):
        cors.add(route)

    return app



import logging
import datetime
import copy
from sqlalchemy.sql import text
from potnanny.utils import iso_from_sqlite
from potnanny.models.room import Room
from potnanny.models.device import Device
from potnanny.models.measurement import Measurement
from potnanny.models.interface import execute_statement


logger = logging.getLogger(__name__)


DEFAULT_GRAPH = {
    'type': 'line',
    'options': {
        'maintainAspectRatio': False,
        'responsive': True,
        'plugins': {
            'title': {
                'display': True,
                'text': None,
            }
        }
    },
    'data': {
        'datasets': []
    }
}


async def room_graph(pk: int, mtype: str, hours: int = 12):
    """
    Get graph data for room
    """

    now = datetime.datetime.utcnow()
    start = str(now - datetime.timedelta(hours=hours))
    graph = copy.deepcopy(DEFAULT_GRAPH);

    sql = """
        SELECT
          d.id, d.name, m.type, m.value, m.created
        FROM
          measurements as m,
          devices as d
        WHERE
            d.room_id = %d
        AND
            m.device_id = d.id
        AND
            m.created >= '%s'
    """ % (pk, start)

    if mtype is not None:
        sql += """
        AND
            m.type = '%s' """ % mtype

    sql += """
        ORDER BY
            m.created """

    data = {}
    labels = []
    results = await execute_statement(text(sql))
    rows = results.all()
    for r in rows:
        if graph['options']['plugins']['title']['text'] is None:
            graph['options']['plugins']['title']['text'] = mtype.upper()

        if r[0] not in data:
            data[r[0]] = {
                'label': r[1],
                'data': [] }
        dt = iso_from_sqlite(r[4])

        if dt not in labels:
            labels.append(dt)

        data[r[0]]['data'].append({
            'y': r[3],
            'x': dt,
        })

    # graph['data']['labels'] = sorted(list(set(labels)))
    graph['data']['labels'] = sorted(labels)
    graph['data']['datasets'] = [v for v in data.values()]

    return graph


async def device_graph(pk: int, hours: int = 12, mtype = None):
    """
    Get graph data for device
    """

    now = datetime.datetime.utcnow()
    start = str(now - datetime.timedelta(hours=hours))
    graph = copy.deepcopy(DEFAULT_GRAPH);

    sql = """
        SELECT
          d.id, d.name, m.type, m.value, m.created
        FROM
          measurements as m,
          devices as d
        WHERE
            d.id = %d
        AND
            m.device_id = d.id
        AND
            m.created >= '%s'
    """ % (pk, start)

    if mtype is not None:
        sql += """
        AND
            m.type = '%s' """ % mtype

    sql += """
        ORDER BY
            m.created """

    data = {}
    labels = []
    results = await execute_statement(text(sql))
    rows = results.all()
    for r in rows:
        if graph['options']['plugins']['title']['text'] is None:
            graph['options']['plugins']['title']['text'] = r[1]

        if r[2] not in data:
            data[r[2]] = {
                'label': r[2],
                'data': [] }
        dt = iso_from_sqlite(r[4])
        if dt not in labels:
            labels.append(dt)

        data[r[2]]['data'].append({
            'y': r[3],
            'x': dt,
        })

    graph['data']['labels'] = sorted(list(set(labels)))
    graph['data']['datasets'] = [v for v in data.values()]
    return graph



import asyncio
import re
import logging
from potnanny.utils.shell import run

logger = logging.getLogger(__name__)
DEFAULT_PATH = "/etc/wpa_supplicant/wpa_supplicant.conf"
DEFAULT_SOCK = "/var/run/wpa_supplicant"


async def current_wifi():
    """
    get name of currently connected wifi
    """

    cmd = 'sudo iwgetid -r'
    try:
        rval, stdout, stderr = await run(cmd);
        if rval == 0:
            network = stdout.decode().strip()
            return network
    except:
        logger.warning(f"unexpected error: {stderr.decode().strip()}")
        pass

    return None


async def wifi_networks() -> list:
    """
    Scan for wifi network info. returns a list of dicts
    """

    results = []
    current = await current_wifi()
    cmd = 'sudo iwlist wlan0 scan'

    try:
        rval, stdout,stderr = await run(cmd)
        if rval != 0:
            logger.warning(f"unexpected error: {stderr.decode().strip()}")
            return []

        raw = stdout.decode().strip()
        results = parse_wifi_results(raw)
        for r in results:
            if 'essid' in r and r['essid'] == current:
                r['in_use'] = True
    except Exception as x:
        logger.warning(str(x))

    return results


async def wpa_networks() -> dict:
    """
    Get wpa supplicant known network data as dict: {id: ssid, ...}
    """

    data = {}
    cmd = "wpa_cli list_networks"
    try:
        rval, stdout, stderr = await run(cmd)
        if proc.returncode != 0:
            logger.warning(stderr.decode().strip())
            return {}

        for line in stdout.decode().splitlines():
            if re.search(r'^(Selected interface|network id)', line):
                continue
            if re.search(r'^\s*$', line):
                continue

            atoms = line.split()
            if len(atoms) >= 2:
                data[atoms[0]] = atoms[1]
    except:
        pass

    return data


async def wpa_drop_network(ssid: str) -> tuple:
    """
    Removed named network from the wpa config
    returns tuple(exit-code, message)
    """

    target = None
    data = await wpa_networks()
    for k,v in data.items():
        if v == ssid:
            target = k
            break

    if target is None:
        return (1, f"network ssid {ssid} not found")

    cmd = f"sudo wpa_cli remove_network {target}"
    try:
        rval, stdout, stderr = await run(cmd)
        if rval != 0:
            logger.warning(stderr.decode())
            return (rval, f"unexpected error: {stderr.decode()}")
    except:
        pass

    return (0, "ok")


async def wpa_password_entry(ssid: str, pw: str) -> str:
    """
    Generate a network entry for wpa_supplicant.conf, from the SSID/password
    """

    cmd = f"wpa_passphrase {ssid} {pw}"
    proc = await asyncio.create_subprocess_shell(
        cmd,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE)

    stdout, stderr = await proc.communicate()
    if proc.returncode != 0:
        logger.warning(stderr.decode())
        return ""

    atoms = stdout.decode().split("\n")

    # the priority above zero is important, for the way potnanny is working
    # with wpa, and the auto AP hotspot function.
    atoms.insert(-3, "\tpriority=10")

    # join to str, but omit the plain text passwd from the output
    return "\n".join([a for a in atoms if not a.startswith("\t#")]).strip()


async def append_to_wpaconf(entry: str, path: str = DEFAULT_PATH):
    """
    Append a network entry block to the wpa_supplicant.conf file
    """

    cmd = f'printf "\n{entry}\n" | sudo tee -a {path}'
    proc = await asyncio.create_subprocess_shell(
        cmd,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE)

    stdout, stderr = await proc.communicate()
    if proc.returncode != 0:
        logger.warning(stderr.decode())
        return (proc.returncode, "unexpected error: {stderr.decode()}")

    return (0, "ok")


def parse_wifi_results(txt: str) -> list:
    results = {}
    bufr = []
    for line in txt.splitlines():
        if re.search(r'Cell \d+', line):
            if bufr:
                cell = extract_network_data(bufr)
                if cell and 'essid' in cell:
                    results[cell['essid']] = cell
                bufr = []
        bufr.append(line)

    # last check of the remaining buffer
    if bufr:
        cell = extract_network_data(bufr)
        if cell and 'essid' in cell:
            results[cell['essid']] = cell

    return list(results.values())


def extract_network_data(bufr: list) -> dict:
    lookups = {
        'address': re.compile(r'Cell \d+ - Address: (\S+)'),
        'channel': re.compile(r'Channel:(\d+)'),
        'essid': re.compile(r'ESSID:\"(\S+)\"'),
        'quality': re.compile(r'Quality=(\d+/\d+)'),
        'signal': re.compile(r'Signal level=(-\d+ dBm)'),
        'ieee': re.compile(r'IE: (IEEE.+)'),
        'authentication': re.compile(r'Authentication Suites \(\d+\) : (.+)'),
        'mode': re.compile(r'Mode:(.+)'),
        'encryption': re.compile(r'Encryption key:(.+)'),
    }
    results = {}

    for line in bufr:
        for key, regex in lookups.items():
            match = regex.search(line)
            if match:
                results[key] = match.group(1).strip()

    return results



import asyncio
import unittest
import datetime
import random
from unittest import IsolatedAsyncioTestCase
from sqlalchemy.sql import text
import potnanny.database as db
from potnanny.models.interface import ObjectInterface, execute_statement
from potnanny.models.room import Room
from potnanny.models.device import Device
from potnanny.models.measurement import Measurement
from potnanny.controllers.environment import get_environments

CREATED = False
NOW = datetime.datetime.now().replace(second=0, microsecond=0)

class TestQueries(IsolatedAsyncioTestCase):
# ###############################################

    async def asyncSetUp(self):
        uri = 'sqlite+aiosqlite://'
        await db.init_db(uri)

        if not CREATED:
            await self.build_models()

    async def build_models(self):
        global CREATED
        CREATED = True

        # add rooms
        for room in [1,2]:
            try:
                opts = {'name': f"Room {room}"}
                r = Room(**opts)
                await r.insert()
            except Exception as x:
                pass

            # add room devices
            for device in [1,2,3]:
                try:
                    opts = {
                        'name': f"Device {device}/{r.id}",
                        'room_id': r.id}
                    d = Device(**opts)
                    await d.insert()
                except Exception as x:
                    pass

                for minute in range(0,30,10):
                    dt = NOW - datetime.timedelta(minutes=minute)
                    for t in ['temperature','humidity','vpd']:
                        v = random.random() * 10
                        opts = {
                            'type': t,
                            'value': v,
                            'device_id': d.id,
                            'created': dt
                        }
                        obj = Measurement(**opts)
                        try:
                            await obj.insert()
                        except Exception as x:
                            pass


    async def asyncTearDown(self):
        pass


    async def test_check_measurements(self):
        stmt = text("""
            SELECT * FROM
                (SELECT id, name FROM rooms) as a
                LEFT JOIN (
                    SELECT
                        d.room_id,
                        m.type,
                        m.value,
                        avg(m.value),
                        max(m.created)
                    FROM
                        measurements as m,
                        devices as d
                    WHERE
                        m.device_id = d.id
                    AND
                        m.type IN ('vpd','temperature', 'humidity')
                    GROUP BY
                        d.room_id,
                        m.type
                ) as b
                ON a.id = b.room_id

        """)
        results = await execute_statement(stmt)
        rows = results.all()
        for r in rows:
            print(r)



if __name__ == '__main__':
    unittest.main()



import logging
from sqlalchemy import (Column, Integer, DateTime, Unicode, UnicodeText,
    Boolean, ForeignKey, func)
from marshmallow import Schema, fields, EXCLUDE, INCLUDE
from potnanny.database import Base
from potnanny.models.mixins import CRUDMixin


logger = logging.getLogger(__name__)


class UserSchema(Schema):
    class Meta:
        unknown = EXCLUDE

    name = fields.String()
    password = fields.String(allow_none=True)
    roles = fields.String(allow_none=True)


class User(Base, CRUDMixin):
    __tablename__ = 'users'

    id = Column(Integer, primary_key=True)
    name = Column(Unicode(128), nullable=False, unique=True)
    password = Column(Unicode(256), nullable=False, unique=False)
    roles = Column(Unicode(256), nullable=False, unique=False, default='user')
    is_active = Column(Boolean, default=True)
    created = Column(DateTime, server_default=func.now())

    # make relationships compatible with asyncio
    __mapper_args__ = {"eager_defaults": True}


    def __repr__(self):
        return "<User(id={}, name={})>".format(self.id, self.name)


    def as_dict(self):
        return {
            'id': self.id,
            'name': self.name,
            'roles': self.roles,
            'created': self.created.isoformat() + "Z",
            'is_active': self.is_active,
        }


    def has_role(self, role):
        if role in self.roles.split(','):
            return True

        return False


    def add_role(self, role):
        if not self.has_role(role):
            if self.roles is None:
                self.roles = role
            else:
                self.roles += ",%s" % role


    def remove_role(self, role):
        atoms = self.roles.split(',')
        try:
            atoms.remove(role)
        except:
            pass

        self.roles = ",".join(atoms)



import logging
from potnanny.locks import LOCKS


logger = logging.getLogger(__name__)


class GPIOInterface:

    def __init__(self, *args, **kwargs):
        self.lock = LOCKS['gpio']
        for k, v in kwargs.items():
            if hasattr(self, k):
                setattr(self, k, v)


    async def poll_devices(self, devices):
        """
        poll measurements from list of devices

        args:
            - list of Device objects
        returns:
            - list of measurements
        """

        results = []

        # filter only gpio devices
        valid = self._gpio_list(devices)

        # run polling and gather results when done
        for d in valid:
            data = None
            try:
                if self.lock:
                    async with self.lock:
                        data = await d.poll()
                else:
                    data = await d.poll()
            except Exception as x:
                logger.warning(x)

            if data:
                results.append(data)

        return results


    def _gpio_list(self, devices):
        # parse list of devices and return only gpio pollable clients
        results = []
        for d in devices:
            if (not hasattr(d.plugin, 'is_gpio') or
                d.plugin.is_bluetooth is not True):
                continue

            if (not hasattr(d.plugin, 'is_pollable') or
                d.plugin.is_pollable is not True):
                continue

            results.append(d)

        return results



import os
import unittest
import datetime
from potnanny.utils.pw import hash_password, verify_password

class TestPWHashUtils(unittest.TestCase):
# ###############################################

    def test_hashing(self):
        h1 = hash_password('foobar123')

        assert verify_password('foobar123', h1) is True
        assert verify_password('foobar123!', h1) is False

if __name__ == '__main__':
    unittest.main()



import re
import asyncio
import logging
import datetime
from sqlalchemy import (Column, Integer, Text, String, Float, Boolean,
    DateTime, ForeignKey, func)
from sqlalchemy.orm import relationship
from potnanny.database import Base
from potnanny.plugins.mixins import InterfaceMixin
from potnanny.controllers.trigger import (open_action_triggers,
    new_action_trigger)
from potnanny.utils import evaluate
from potnanny.models.mixins import CRUDMixin
from potnanny.models.ext import MutableDict, JSONEncodedDict


logger = logging.getLogger(__name__)


class Action(Base, CRUDMixin, InterfaceMixin):
    __tablename__ = 'actions'

    id = Column(Integer, primary_key=True)
    name = Column(String(64), nullable=False, unique=False)
    interface = Column(String(128), nullable=True, unique=False)
    created = Column(DateTime, server_default=func.now())
    attributes = Column(MutableDict.as_mutable(JSONEncodedDict), nullable=True)

    # make relationships compatible with asyncio
    __mapper_args__ = {"eager_defaults": True}

    # relationships
    device_id = Column(Integer, ForeignKey('devices.id'))


    def __repr__(self):
        return "<Action ({})>".format(self.name)


    def as_dict(self):
        data = {
            'id': self.id,
            'name': self.name,
            'created': self.created.isoformat() + "Z",
            'device_id': self.device_id,
            'interface': self.interface,
            'attributes': self.attributes,
        }

        return data


    async def _existing_trigger(self):

        try:
            minutes = int(self.attributes['sleep_minutes'])
            now = datetime.datetime.utcnow()
            start = now - datetime.timedelta(minutes=minutes)
            results = await open_action_triggers(self.id, start)
            if results:
                return True
        except:
            pass

        return False


    async def input(self, data):
        """
        Accept measurement data as input, evaluate, and perform actions
        as required:
        args:
            - dict (measurement data from a device)
        returns:
        """

        success = False

        if self._precheck(data) is not True:
            logger.debug(f"data {data} is not acceptable. skipping")
            return

        try:
            equation = "%0.1f %s %s" % (
                float(data['value']),
                self.attributes['condition'],
                self.attributes['threshold'])
            if evaluate(equation) is not True:
                return
        except:
            return

        if await self._existing_trigger():
            logger.debug(f"data {data} has an existing trigger. skipping")
            return

        try:
            klass = self.interface_class(self.interface)
            obj = klass(action_id=self.id, action_name=self.name)
            success = await obj.input(data)
        except Exception as x:
            logger.debug(f"Action interface plugin fail: {x}")
            return

        try:
            if (success and 'sleep_minutes' in self.attributes and
                self.attributes['sleep_minutes'] > 0):
                await new_action_trigger(self.id)
        except Exception as x:
            logger.warning(f"Failed to create action trigger for {self.name}")


    def _precheck(self, data):
        """
        examine incoming data, decide if it is valid for us
        """

        try:
            if data['type'] != self.attributes['type']:
                return False
        except:
            return False

        try:
            if data['device_id'] != self.device_id:
                return False
        except:
            return False

        return True



import logging
import datetime
from aiohttp import web
from potnanny.utils.wifi import (wifi_networks, append_to_wpaconf,
    wpa_password_entry)
from .decorators import login_required

routes = web.RouteTableDef()
logger = logging.getLogger(__name__)

@routes.get('/api/v1.0/wifi/networks')
@login_required
async def get_networks(request):
    results = await wifi_networks()
    if not results:
        return web.json_response({
            "status": "error", "msg": "no data"}, status=404)

    return web.json_response({
        "status": "ok", "msg": results}, status=200)


@routes.post('/api/v1.0/wifi/join')
@login_required
async def join_network(request):
    data = await request.json()
    entry = await wpa_password_entry(data['ssid'], data['password'])
    rval, msg = await append_to_wpaconf(entry)

    if rval != 0:
        return web.json_response({
            "status": "error", "msg": f"error joining network: {msg}"},
            status=409)

    return web.json_response({
        "status": "ok", "msg": "reboot raspberry pi to join the new network"},
        status=201)



import os

def resolve_path(path):
    """
    expand path to fully qualified, absolute path name
    args:
        - str (pathname)
    returns:
        str (path)
    """

    return os.path.abspath(os.path.expanduser(path))



import logging
from marshmallow import Schema, fields, EXCLUDE, INCLUDE
from sqlalchemy import (Column, Integer, String, Float, DateTime,
    ForeignKey, PickleType, Text, Boolean, func)
from potnanny.database import Base
from potnanny.models.mixins import CRUDMixin
from potnanny.models.ext import MutableDict, JSONEncodedDict


logger = logging.getLogger(__name__)


class KeychainSchema(Schema):
    class Meta:
        unknown = EXCLUDE

    name = fields.String()
    attributes = fields.Dict(allow_none=True)
    protected = fields.Boolean(allow_none=True)


class Keychain(Base, CRUDMixin):
    """Store named info in a key-value pair."""

    __tablename__ = 'keychains'

    id = Column(Integer, primary_key=True)
    name = Column(String(32), nullable=False, unique=True)
    attributes = Column(MutableDict.as_mutable(JSONEncodedDict), nullable=False)
    protected = Column(Boolean, nullable=False, default=False)
    created = Column(DateTime, server_default=func.now())
    modified = Column(DateTime, server_default=func.now())

    __mapper_args__ = {'eager_defaults': True}


    def __repr__(self):
        return "<Keychain({})>".format(self.name)


    def as_dict(self):
        data = {
            'id': self.id,
            'name': self.name,
            'attributes': self.attributes,
            'protected': self.protected }
        return data



import asyncio
import unittest
import datetime
import random
from unittest import IsolatedAsyncioTestCase
import potnanny.database as db
from potnanny.models.interface import ObjectInterface
from potnanny.models.room import Room
from potnanny.models.device import Device
from potnanny.models.measurement import Measurement
from potnanny.controllers.environment import (get_environments,
    get_room_environments)

# globals
CREATED = False
NOW = datetime.datetime.now().replace(second=0, microsecond=0)

class TestEnvironmentCtrl(IsolatedAsyncioTestCase):
# ###############################################

    async def asyncSetUp(self):
        uri = 'sqlite+aiosqlite://'
        await db.init_db(uri)

        if not CREATED:
            await self.build_models()

    async def build_models(self):
        global CREATED
        CREATED = True

        # add rooms
        for room in [1,2]:
            try:
                opts = {'name': f"Room {room}"}
                r = Room(**opts)
                await r.insert()
            except Exception as x:
                pass

            # add room devices
            for device in [1,2,3]:
                try:
                    opts = {
                        'name': f"Device {device}/{r.id}",
                        'room_id': r.id}
                    d = Device(**opts)
                    await d.insert()
                except Exception as x:
                    pass

                start = 0
                for minute in range(0,30,10):
                    dt = NOW - datetime.timedelta(minutes=minute)
                    for t in ['temperature','humidity','vpd', 'battery']:
                        if t == 'temperature':
                            if not start:
                                start = 60
                            v = start + minute
                        else:
                            v = random.random() * 10

                        opts = {
                            'type': t,
                            'value': v,
                            'device_id': d.id,
                            'created': dt
                        }
                        obj = Measurement(**opts)
                        try:
                            await obj.insert()
                        except Exception as x:
                            pass


    async def asyncTearDown(self):
        pass


    async def test_get_environment(self):
        results = await get_environments()
        assert len(results) == 2


    async def test_room_environment(self):
        results = await get_room_environments(2)
        print(results)
        assert len(results) == 3



if __name__ == '__main__':
    unittest.main()



import asyncio
import unittest
import datetime
from unittest import IsolatedAsyncioTestCase
import potnanny.database as db
from potnanny.models.measurement import Measurement
from potnanny.controllers.outlet import OutletControl


class TestOutletCtrl(IsolatedAsyncioTestCase):
# ###############################################

    async def asyncSetUp(self):
        uri = 'sqlite+aiosqlite://'
        await db.init_db(uri)

        now = datetime.datetime.utcnow().replace(second=0, microsecond=0)
        for outlet in [1, 2]:
            for i in range(1,10):
                then = now - datetime.timedelta(minutes=i)
                opts = {
                    'device_id': 1,
                    'type': f"outlet_{outlet}",
                    'value': outlet - 1,
                    'created': then
                }
                m = Measurement(**opts)
                await m.insert()

    async def asyncTearDown(self):
        pass

    async def test_check_outlet1(self):
        ctl = OutletControl(1, 1)
        state1 = await ctl.last_state(90)
        assert state1 == 0

        state2 = await ctl.last_state(20)
        assert state2 == -1


    async def test_check_outlet2(self):
        ctl = OutletControl(1, 2)
        state1 = await ctl.last_state(90)
        assert state1 == 1

        state2 = await ctl.last_state(20)
        assert state2 == -1


if __name__ == '__main__':
    unittest.main()



import os
import re
import sys
import glob
import logging
import importlib
import types
from potnanny.utils import resolve_path

logger = logging.getLogger(__name__)

def load_plugins(path):
    """
    Load plugins from named path
    """

    files = []
    forbidden = ('.', '_', '__init__.py')

    path = resolve_path(path)
    logger.debug("Loading plugins from %s" % path)
    if not os.path.exists(path):
        raise IOError("Plugin path '%s' does not exist" % path)

    for dir, subdirs, flist in os.walk(path):
        subdirs[:] = [ d for d in subdirs if not d.startswith(forbidden) ]

        for f in flist:
            fpath = os.path.join(dir, f)
            if f.endswith('.py') and f not in forbidden:
                files.append(fpath)

    for f in files:
        try:
            # build plug module name like 'input.ble.govee.Hygrometer'
            tree = re.split(r'\\|/', re.sub(path, '', f))
            tree[-1] = tree[-1].split('.')[0]
            mod_name = '.'.join(tree[1:])
            loader = importlib.machinery.SourceFileLoader(mod_name, f)
            mod = types.ModuleType(loader.name)
            loader.exec_module(mod)
            logger.info("Loaded plugin %s" % mod_name)
        except Exception as x:
            logger.warning(str(x))



import os

PIDFILE = os.path.expanduser('~/.potnanny.pid')


def is_running(path=PIDFILE):
    """
    Check if an instance of the app is already running.
    args:
        - path to pidfile
    returns:
        - pid of running program, or 0 if not running.
    """

    if os.path.exists(path):
        st = os.stat(path)
        if st.st_size > 0:
            pid = pid_from_file(path)
            if check_pid(pid) is True:
                return pid
            else:
                os.unlink(path)
        else:
            os.unlink(path)

    return 0


def pid_from_file(path=PIDFILE):
    """
    Get pid number from named file
    args:
        - path to filename
    returns:
        - int (pid number)
    """

    with open(path, 'r') as fh:
        return int(fh.read())


def pid_to_file(pid, path=PIDFILE):
    """
    Write a pid to named file
    args:
        - pid number
        - path to filename
    """

    with open(path, 'w') as fh:
        fh.write(f"{pid}\n")


def check_pid(pid: int):
    """
    Check For the existence of a running unix pid
    args:
        - pid number
    returns:
        bool, running pid was found = True
    """

    try:
        os.kill(pid, 0)
    except OSError:
        return False

    return True



import os
import unittest
import datetime
from potnanny.utils.vpd import calculate_vpd

class TestUtilFuncs(unittest.TestCase):
# ###############################################

    def test_calc(self):
        result = calculate_vpd(24.0, 55.0)
        assert result == 1.34


if __name__ == '__main__':
    unittest.main()



import logging
import datetime
import potnanny.database as db
from sqlalchemy.sql.expression import select
from sqlalchemy import delete
from potnanny.locks import LOCKS


logger = logging.getLogger(__name__)


async def execute_statement(stmt):
    results = None
    try:
        async with db.session() as session:
            results = await session.execute(stmt)
    except Exception as x:
        logger.warning(x)
        pass

    return results


class ObjectInterface:
    """
    Class to query Sqlalchemy object models, on the most common ways to fetch
    them:
        - by id
        - by name
        - all of them
    """

    def __init__(self, target_class):
        self.klass = target_class


    async def get_by_id(self, obj_id:int):
        """
        Get an object by its ID attribute
        """

        stmt = select(self.klass).filter(self.klass.id == obj_id)
        try:
            results = await execute_statement(stmt)
            obj = results.one_or_none()[0]
            logger.debug(f"Object: {obj}")
            return obj
        except:
            return None


    async def get_by_name(self, name:str):
        """
        Get an object by its NAME attribute
        """

        stmt = select(self.klass).filter(self.klass.name == name)
        try:
            results = await execute_statement(stmt)
            obj = results.one_or_none()[0]
            logger.debug(f"Object: {obj}")
            return obj
        except:
            return None


    async def get_all(self):
        """
        Get all objects of the Class
        """

        stmt = select(self.klass)
        try:
            results = await execute_statement(stmt)
            items = [item[0] for item in results.all()]
            return items
        except:
            return []


    async def update(self, pk, data):
        """
        Update data for an object id
        """

        stmt = select(self.klass).filter(self.klass.id == pk)
        obj = None

        async def process(stmt):
            async with db.session() as session:
                changed = 0
                results = await session.execute(stmt)
                obj = results.one_or_none()[0]
                for k, v in data.items():
                    if hasattr(obj, k) and getattr(obj, k) != v:
                        setattr(obj, k, v)
                        changed += 1

                if changed:
                    await session.commit()

        try:
            if 'db' in LOCKS and LOCKS['db'] is not None:
                async with LOCKS['db']:
                    await process(stmt)
            else:
                await process(stmt)

        except Exception as x:
            logger.warning(x)

        return obj


    async def delete(self, pk):
        """
        Delete an object id
        """

        stmt = delete(self.klass).where(self.klass.id == pk)

        async def process(stmt):
            async with db.session() as session:
                results = await session.execute(stmt)
                await session.commit()

        try:
            if 'db' in LOCKS and LOCKS['db'] is not None:
                async with LOCKS['db']:
                    await process(stmt)
            else:
                await process(stmt)

        except Exception as x:
            logger.warning(x)

        return None



from .base import (DevicePlugin, BluetoothDevicePlugin, GPIODevicePlugin, 
    ActionPlugin, PipelinePlugin)

plugin_category_map = {
    'device': [DevicePlugin, BluetoothDevicePlugin, GPIODevicePlugin],
    'action': ActionPlugin,
    'pipeline': PipelinePlugin,
}



import asyncio
import logging
from sqlalchemy import delete
from potnanny.models.interface import execute_statement
from potnanny.models.measurement import Measurement
from potnanny.models.error import Error


logger = logging.getLogger(__name__)


async def purge_measurements(cutoff):
    """
    Remove measurements from table when older than cutoff datetime
    """

    try:
        stmt = (delete(Measurement).where(Measurement.created < cutoff))
        result = await execute_statement(stmt)
    except Exception as x:
        logger.warning(x)

    try:
        stmt = (delete(Error).where(Error.created < cutoff))
        result = await execute_statement(stmt)
    except Exception as x:
        logger.warning(x)



import asyncio
import datetime
from sqlalchemy.sql.expression import select
from potnanny.models.interface import ObjectInterface, execute_statement
from potnanny.models.trigger import Trigger


async def open_action_triggers(action_id: int, start: datetime.datetime):
    """
    Find any open triggers for an action, since the starting time
    """

    stmt = select(Trigger).filter(
        Trigger.action_id == action_id
    ).filter(
        Trigger.created > start
    )

    results = await execute_statement(stmt)
    return results.all()


async def new_action_trigger(action_id: int, message: str = ""):
    """
    Creat a new trigger for an action
    """

    now = datetime.datetime.utcnow().replace(microsecond=0)
    t = Trigger(
        action_id=action_id,
        message=message,
        created=now,
    )
    await t.insert()



import asyncio
import unittest
from unittest import IsolatedAsyncioTestCase
import potnanny.database as db
from potnanny.plugins.mixins import InterfaceMixin
from potnanny.plugins import ActionPlugin


class DeviceTest(ActionPlugin):
    pass


class TestMixins(IsolatedAsyncioTestCase):
# ###############################################

    async def asyncSetUp(self):
        uri = 'sqlite+aiosqlite://'
        await db.init_db(uri)

    async def asyncTearDown(self):
        pass

    async def test_find(self):
        name = '__main__.DeviceTest'
        klass = InterfaceMixin.class_from_name(name)
        obj = klass()
        assert type(obj) is DeviceTest




if __name__ == '__main__':
    unittest.main()



import os
import unittest
import datetime
from potnanny.utils.temperature import convert_to_fahrenheit

class TestUtilFuncs(unittest.TestCase):
# ###############################################

    def test_convert(self):
        assert convert_to_fahrenheit(0) == 32.0
        assert convert_to_fahrenheit(100) == 212.0

    def test_bogus(self):
        assert convert_to_fahrenheit("foo") == None


if __name__ == '__main__':
    unittest.main()



import asyncio
import unittest
import datetime
import json
import potnanny.database as db
from unittest import IsolatedAsyncioTestCase
from potnanny.models.user import User
from potnanny.models.interface import ObjectInterface


class TestUtilsModel(IsolatedAsyncioTestCase):
# ###############################################

    async def asyncSetUp(self):
        uri = 'sqlite+aiosqlite://'
        await db.init_db(uri)

        try:
            opts = {
                'name': 'test',
                'roles': 'user',
                'password': 'invalid!'
            }
            user = User(**opts)
            await user.insert()
        except:
            pass

    async def asyncTearDown(self):
        pass

    async def test_get_name(self):
        ifc = ObjectInterface(User)
        obj = await ifc.get_by_name('test')
        assert obj is not None
        assert obj.name == 'test'

    async def test_get_id(self):
        ifc = ObjectInterface(User)
        obj = await ifc.get_by_id(2)
        assert obj is not None
        assert obj.name == 'test'

    async def test_get_bad_id(self):
        ifc = ObjectInterface(User)
        obj = await ifc.get_by_id(999)
        assert obj is None

    async def test_get_all(self):
        ifc = ObjectInterface(User)
        objects = await ifc.get_all()
        assert len(objects) > 0

    async def test_delete_user(self):
        opts = {
            'name': 'temporary',
            'roles': 'user',
            'password': 'invalid!'}

        user = User(**opts)
        await user.insert()

        pk = user.id

        results = await ObjectInterface(User).delete(pk)
        assert results is None

        obj = await ObjectInterface(User).get_by_id(pk)
        assert obj is None


if __name__ == '__main__':
    unittest.main()



class PluginBase(type):
    """
    This class is the base for all plugins
    """

    def __init__(cls, name, bases, attrs):
        if not hasattr(cls, 'plugins'):
            cls.plugins = []
        else:
            if cls not in cls.plugins:
                cls.plugins.append(cls)

class PluginFinder:
    @classmethod
    def get_named_class(cls, name):
        for p in cls.plugins:
            klass = '.'.join((p.__module__, p.__name__))
            if klass == name:
                return p

        return None


class DevicePlugin(PluginFinder, metaclass=PluginBase):
    # DEPRACATED class, Jan 2023
    pass

class BluetoothDevicePlugin(PluginFinder, metaclass=PluginBase):
    pass

class GPIODevicePlugin(PluginFinder, metaclass=PluginBase):
    pass

class ActionPlugin(PluginFinder, metaclass=PluginBase):
    pass

class PipelinePlugin(PluginFinder, metaclass=PluginBase):
    pass




from functools import wraps
from aiohttp import web
from aiohttp_session import get_session

def login_required(func):
    @wraps(func)
    async def decorated(request):
        session = await get_session(request)
        if not session or session.new is True or 'username' not in session:
            return web.json_response({
                "status": "error", "msg": "session not authenticated"},
                status=401)

        return await func(request)

    return decorated



import logging
from aiohttp import web
from potnanny.models.schedule import Schedule, ScheduleSchema
from potnanny.models.interface import ObjectInterface
from .decorators import login_required

routes = web.RouteTableDef()
logger = logging.getLogger(__name__)

@routes.get('/api/v1.0/schedules')
@login_required
async def get_list(request):
    params = request.rel_url.query
    objects = await ObjectInterface(Schedule).get_all()
    if not objects:
        return web.json_response({
            "status": "error", "msg": "no data"}, status=404)

    if params and 'device' in params:
        if 'outlet' in params:
            payload = [ o.as_dict() for o in objects if (
                o.device_id == int(params['device']) and
                o.outlet == int(params['outlet'])) ]
        else:
            payload = [
                o.as_dict() for o in objects if (
                o.device_id == int(params['device'])) ]
    else:
        payload = [
            o.as_dict() for o in objects]

    return web.json_response({
        "status": "ok", "msg": payload}, status=200)


@routes.post('/api/v1.0/schedules')
@login_required
async def create(request):
    jsondata = await request.json()
    schema = ScheduleSchema()
    data = schema.load(jsondata)

    try:
        obj = Schedule(**data)
        await obj.insert()
        return web.json_response({
            "status": "ok", "msg": obj.as_dict()}, status=201)
    except Exception as x:
        return web.json_response({
            "status": "error", "msg": str(x) }, status=500)


@routes.get('/api/v1.0/schedules/{pk}')
@login_required
async def get(request):
    payload = None
    pk = int(request.match_info['pk'])
    obj = await ObjectInterface(Schedule).get_by_id(pk)
    if not obj:
        return web.json_response({
            "status": "error", "msg": f"object id {pk} not found"},
            status=404)

    return web.json_response({
        "status": "ok", "msg": obj.as_dict()}, status=200)


@routes.patch('/api/v1.0/schedules/{pk}')
@login_required
async def patch(request):
    changed = 0
    pk = int(request.match_info['pk'])
    jsondata = await request.json()
    schema = DeviceSchema()
    data = schema.load(jsondata)

    obj = await ObjectInterface(Schedule).get_by_id(pk)
    if not obj:
        return web.json_response({
            "status": "error", "msg": f"object with id {pk} not found"},
            status=404)

    for k, v in data.items():
        if hasattr(obj, k) and getattr(obj, k) != v:
            setattr(obj, k, v)
            changed += 1

    if changed:
        try:
            await obj.update()
        except Exception as x:
            return web.json_response({
                "status": "error", "msg": str(x) }, status=500)

    return web.json_response({
        "status": "ok", "msg": obj.as_dict()}, status=200)


@routes.delete('/api/v1.0/schedules/{pk}')
@login_required
async def delete(request):
    pk = int(request.match_info['pk'])
    await ObjectInterface(Schedule).delete(pk)
    return web.json_response({
            "status": "ok", "msg": ""}, status=200)



import datetime
import logging
from aiohttp import web
from potnanny.models.keychain import Keychain, KeychainSchema
from potnanny.models.interface import ObjectInterface
from .decorators import login_required

routes = web.RouteTableDef()
logger = logging.getLogger(__name__)

@routes.get('/api/v1.0/keychains')
@login_required
async def get_list(request):
    objects = await ObjectInterface(Keychain).get_all()

    if not objects:
        return web.json_response({
            "status": "error", "msg": "no data"}, status=404)

    return web.json_response({
        "status": "ok", "msg": [o.as_dict() for o in objects]}, status=200)


@routes.post('/api/v1.0/keychains')
@login_required
async def create(request):
    jsondata = await request.json()
    schema = KeychainSchema()
    data = schema.load(jsondata)

    try:
        obj = Keychain(**data)
        await obj.insert()
        return web.json_response({
            "status": "ok", "msg": obj.as_dict()}, status=201)
    except Exception as x:
        logger.warning(x)
        return web.json_response({
            "status": "error", "msg": str(x) }, status=500)


@routes.get('/api/v1.0/keychains/{pk}')
@login_required
async def get(request):
    pk = int(request.match_info['pk'])
    obj = await ObjectInterface(Keychain).get_by_id(pk)
    if not obj:
        return web.json_response({
            "status": "error", "msg": f"object with id {pk} not found"},
            status=404)

    return web.json_response({
        "status": "ok", "msg": obj.as_dict()}, status=200)


@routes.patch('/api/v1.0/keychains/{pk}')
@login_required
async def patch(request):
    pk = int(request.match_info['pk'])
    jsondata = await request.json()
    schema = KeychainSchema()
    data = schema.load(jsondata)

    obj = await ObjectInterface(Keychain).update(pk, data)
    if not obj:
        return web.json_response({
            "status": "error", "msg": f"object with id {pk} not found"},
            status=404)

    return web.json_response({
        "status": "ok", "msg": obj.as_dict()}, status=200)


@routes.delete('/api/v1.0/keychains/{pk}')
@login_required
async def delete(request):
    pk = int(request.match_info['pk'])
    obj = await ObjectInterface(Keychain).get_by_id(pk)
    if not obj:
        return web.json_response({
            "status": "error", "msg": f"object with id {pk} not found"},
            status=404)

    if obj.protected is True:
        return web.json_response({
            "status": "error",
            "msg": "A protected keychain cannot be removed."
        }, status=404)

    await ObjectInterface(Keychain).delete(pk)
    return web.json_response({
        "status": "ok", "msg": ""}, status=200)



import os
import unittest
import datetime
from potnanny.utils.times import (utcnow, datetime_from_iso, 
    datetime_from_sqlite, iso_from_sqlite)


class TestUtilFuncs(unittest.TestCase):
# ###############################################

    def test_utcnow(self):
        dt = utcnow()
        assert type(dt) == datetime.datetime

    def test_conversion(self):
        dt = datetime_from_iso("1970-01-01T00:00:00")
        assert type(dt) == datetime.datetime

    def test_sqlite(self):
        d = '2021-03-14 00:24:41.123'
        dt = datetime_from_sqlite(d)
        assert type(dt) == datetime.datetime

    def test_sqlite_iso(self):
        d = '2021-03-14 00:24:41.123'
        iso = iso_from_sqlite(d)
        assert iso == '2021-03-14T00:24:41.123000Z'

if __name__ == '__main__':
    unittest.main()



import logging
from aiohttp import web
from potnanny.utils.shell import run
from potnanny.controllers.discover import discover_new_devices
from potnanny.controllers.worker import restart_worker
from potnanny.utils.serial import load_serial_number
from potnanny.utils.network import aio_has_www
from .decorators import login_required

routes = web.RouteTableDef()
logger = logging.getLogger(__name__)

@routes.get('/api/v1.0/ping')
async def ping(request):
    return web.json_response({
        "status": "ok", "msg": "pong"}, status=200)


@routes.get('/api/v1.0/scan')
@login_required
async def scan(request):
    results = await discover_new_devices()
    return web.json_response({
        "status": "ok", "msg": results}, status=200)


@routes.post('/api/v1.0/reboot')
@login_required
async def reboot(request):
    (rval, stdout, stderr) = await run('sudo reboot now')


@routes.get('/api/v1.0/serial')
@login_required
async def serial_number(request):
    sn = await load_serial_number()
    return web.json_response({
            "status": "ok", "msg": sn }, status=200)


@routes.get('/api/v1.0/internet')
@login_required
async def has_internet(request):
    success = await aio_has_www()
    if success:
        return web.json_response({
                "status": "ok", "msg": "internet connection" }, status=200)
    else:
        return web.json_response({
                "status": "ok", "msg": "no internet connection" }, status=404)



import asyncio
import logging

logger = logging.getLogger(__name__)


LOCKS = {
    'db':           None,       # global database lock
    'bluetooth':    None,       # global bluetooth lock
    'gpio':         None,       # global gpio lock (future)
}


async def init_locks():
    """
    Initialize locks required for shared asyncio access to resources

    args: none
    returns: none
    """

    logger.debug("Initializing asyncio locks")

    if LOCKS['db'] is None:
        LOCKS['db'] = asyncio.Lock()

    if LOCKS['bluetooth'] is None:
        LOCKS['bluetooth'] = asyncio.Lock()

    if LOCKS['gpio'] is None:
        LOCKS['gpio'] = asyncio.Lock()



import os
import logging
from sqlalchemy.ext.asyncio import (create_async_engine, AsyncSession,
    AsyncAttrs)
from sqlalchemy.orm import DeclarativeBase
from sqlalchemy.orm import sessionmaker
from potnanny.utils import hash_password


engine = None
session = None
logger = logging.getLogger(__name__)


class Base(AsyncAttrs, DeclarativeBase):
    pass


async def init_db(uri, **kwargs):
    """
    Create db engine, sessionmaker, and tables.
    args:
        - uri, str
    returns:
    """

    if not engine:
        await init_engine(uri, **kwargs)

    if not session:
        await init_sessionmaker(engine)

    await init_tables()


async def init_engine(uri, **kwargs):
    """
    Init the engine.

    args:
        - uri, sqlalchemy engine db uri init string
        - optional keyword args to pass to sqlalchemy create_async_engine func
    returns:
    """

    global engine
    logger.debug(f"Initializing db engine {uri}")
    engine = create_async_engine(uri, **kwargs)


async def init_sessionmaker(engine=engine):
    """
    Init the sessionmaker function with the engine.

    args:
        - an engine instance
    returns:
    """

    global session
    logger.debug("Initializing db session")
    session = sessionmaker(engine,
        expire_on_commit=False, class_=AsyncSession)


async def init_tables():
    """
    Create tables based on sqlalchemy models.

    args:
    returns:
    """

    logger.debug("Importing models")

    # import the  models
    from potnanny.models.keychain import Keychain
    from potnanny.models.user import User
    from potnanny.models.room import Room
    from potnanny.models.device import Device
    from potnanny.models.measurement import Measurement
    from potnanny.models.control import Control
    from potnanny.models.schedule import Schedule
    from potnanny.models.error import Error
    from potnanny.models.action import Action
    from potnanny.models.trigger import Trigger

    # create tables
    async with engine.begin() as conn:
        await conn.run_sync(Base.metadata.create_all)

    await init_features()


async def init_features():
    # create initial default config
    try:
        attrs = {
            'temperature_display': 'F',
            'polling_interval': 10,
            'plugin_path': os.path.expanduser('~/potnanny/plugins'),
            'leaf_offset': -2,
            'storage_days': 5,
        }
        kc = Keychain(name='settings', attributes=attrs, protected=True)
        await kc.insert()
    except:
        pass

    # create the default admin user
    try:
        attrs = {
            'name': 'admin',
            'roles': 'admin,user',
            'password': hash_password('potnanny!'),
        }
        user = User(**attrs)
        await user.insert()
    except:
        pass

    try:
        opts = {
            'name': 'features',
            'protected': True,
            'attributes': {
                'room_limit': 1,
                'device_limit': 4
            }
        }
        kc = Keychain(**opts)
        await kc.insert()
    except:
        pass



from aiohttp import web
from potnanny.models.keychain import Keychain
from potnanny.models.interface import ObjectInterface
from marshmallow import Schema, fields, validate, EXCLUDE
from .decorators import login_required

routes = web.RouteTableDef()

class SettingSchema(Schema):
    class Meta:
        unknown = EXCLUDE

    temperature_display = fields.String(allow_none=True,
        validate=validate.OneOf(['f','F','c','C']))
    polling_interval = fields.Integer(allow_none=True,
        validate=validate.Range(min=1,max=30))
    plugin_path = fields.String(allow_none=True)
    leaf_offset = fields.Integer(allow_none=True,
        validate=validate.Range(min=-5,max=5))
    storage_days = fields.Integer(allow_none=True,
        validate=validate.Range(min=1,max=365))


@routes.get('/api/v1.0/settings')
@login_required
async def get(request):
    obj = await ObjectInterface(Keychain).get_by_name('settings')
    if not obj:
        return web.json_response({
            "status": "error", "msg": "settings object not found"},
            status=404)

    return web.json_response({
        "status": "ok", "msg": obj.as_dict() }, status=200)


@routes.patch('/api/v1.0/settings')
@login_required
async def patch(request):
    jsondata = await request.json()
    schema = SettingSchema()
    data = schema.load(jsondata['attributes'])
    obj = await ObjectInterface(Keychain).get_by_name('settings')
    if not obj:
        return web.json_response({
            "status": "error", "msg": "settings object not found"},
            status=404)

    pk = obj.id
    shiny = await ObjectInterface(Keychain).update(pk, data)
    if not shiny:
        return web.json_response({
            "status": "error", "msg": "settings object not found"},
            status=404)

    return web.json_response({
        "status": "ok", "msg": shiny.as_dict()}, status=200)



import os
import unittest
from potnanny.utils.lists import flatten_list


class TestUtilFuncs(unittest.TestCase):
# ###############################################

    def test_flatten(self):
        orig = ['foo', 'bar', ['baz', 'banana', ['apple']]]
        result = flatten_list(orig)
        assert len(result) == 5

if __name__ == '__main__':
    unittest.main()



import asyncio
import unittest
from unittest import IsolatedAsyncioTestCase
import potnanny.database as db
from potnanny.models.interface import ObjectInterface
from potnanny.models.user import User


class TestUserModel(IsolatedAsyncioTestCase):
# ###############################################

    async def asyncSetUp(self):
        uri = 'sqlite+aiosqlite://'
        await db.init_db(uri)

    async def asyncTearDown(self):
        pass


    async def test_create(self):
        opts = {
            'name': 'testuser',
            'password': 'qwerty123456',
            'roles': 'user',
        }
        user = User(**opts)

        await user.insert()
        assert user.id > 0
        assert user.name == 'testuser'


    async def test_get_id(self):
        user = await ObjectInterface(User).get_by_id(1)
        assert user is not None


    async def test_get_name(self):
        user = await ObjectInterface(User).get_by_name('admin')
        assert user is not None


    async def test_update(self):
        opts = {
            'name': 'testuser2',
            'password': 'qwerty123456',
            'roles': 'user',
        }
        user = User(**opts)
        await user.insert()

        user.name = 'modified'
        await user.update()
        assert user.name == 'modified'


    async def test_delete(self):
        opts = {
            'name': 'testuser3',
            'password': 'qwerty123456',
            'roles': 'user',
        }
        user = User(**opts)
        await user.insert()
        assert user.id > 0

        pk = user.id
        await user.delete()

        obj = await ObjectInterface(User).get_by_id(pk)
        assert obj is None


if __name__ == '__main__':
    unittest.main()



from .shell import run

SERIAL_NUMBER = None

async def load_serial_number():
    global SERIAL_NUMBER
    if not SERIAL_NUMBER:
        cmd = "cat /proc/cpuinfo | grep Serial | cut -d ' ' -f 2"
        (rval, stdout, stderr) = await run(cmd)
        if rval == 0:
            SERIAL_NUMBER = stdout.decode().strip()

    return SERIAL_NUMBER



import os
import logging
import datetime
import asyncio
from aiohttp import web
from potnanny.plugins import (BluetoothDevicePlugin, GPIODevicePlugin,
    PipelinePlugin, ActionPlugin)
from .decorators import login_required

routes = web.RouteTableDef()
logger = logging.getLogger(__name__)

@routes.get('/api/v1.0/plugins')
@login_required
async def get_plugins(request):
    results = {}

    try:
        results['bluetooth'] = [{
            'name': p.name,
            'description': p.description,
            'interface': '.'.join((p.__module__, p.__name__))
        } for p in BluetoothDevicePlugin.plugins]
    except:
        pass

    try:
        results['gpio'] = [{
            'name': p.name,
            'description': p.description,
            'interface': '.'.join((p.__module__, p.__name__))
        } for p in GPIODevicePlugin.plugins]
    except:
        pass

    try:
        results['pipeline'] = [{
            'name': p.name,
            'description': p.description,
            'interface': '.'.join((p.__module__, p.__name__))
        } for p in PipelinePlugin.plugins]
    except:
        pass

    try:
        results['action'] = [{
            'name': p.name,
            'description': p.description,
            'interface': '.'.join((p.__module__, p.__name__))
        } for p in ActionPlugin.plugins]
    except:
        pass

    return web.json_response({
        "status": "ok", "msg": results}, status=200)



import logging
import json
from sqlalchemy.types import TypeDecorator, VARCHAR
from sqlalchemy.ext.mutable import Mutable

# globals #
logger = logging.getLogger(__name__)


class JSONEncodedDict(TypeDecorator):
    "Represents an immutable structure as a json-encoded string."

    impl = VARCHAR

    def process_bind_param(self, value, dialect):
        if value is not None:
            value = json.dumps(value)
        return value

    def process_result_value(self, value, dialect):
        if value is not None:
            value = json.loads(value)
        return value


class MutableDict(Mutable, dict):
    @classmethod
    def coerce(cls, key, value):
        "Convert plain dictionaries to MutableDict."

        if not isinstance(value, MutableDict):
            if isinstance(value, dict):
                return MutableDict(value)

            # this call will raise ValueError
            return Mutable.coerce(key, value)
        else:
            return value


    def __setitem__(self, key, value):
        "Detect dictionary set events and emit change events."

        dict.__setitem__(self, key, value)
        self.changed()


    def __delitem__(self, key):
        "Detect dictionary del events and emit change events."

        dict.__delitem__(self, key)
        self.changed()



import requests
import asyncio
import aiohttp

def has_www(url='http://www.example.com'):
    """
    Check for internet connectivity
    args:
        - url (optional, defaults to google.com)
    returns:
        - bool
    """

    try:
        resp = requests.get(url)
        if resp.status_code == 200:
            return True
    except:
        pass

    return False


async def aio_has_www(url='http://www.example.com'):
    """
    Asyncio compatible way to check for internet connectivity
    args:
        - url (optional, defaults to google.com)
    returns:
        - bool
    """

    success = False
    async with aiohttp.ClientSession() as session:
        async with session.get(url) as resp:
            if resp.status == 200:
                success = True

    return success



import asyncio
import unittest
import random
import potnanny.database as db
from unittest import IsolatedAsyncioTestCase
from potnanny.models.device import Device
from potnanny.models.measurement import Measurement
from potnanny.models.interface import ObjectInterface
from potnanny.plugins.base import BluetoothDevicePlugin


class MyClass(BluetoothDevicePlugin):
    name = "test"
    description = "testing desc"
    reports = ['temperature', 'humidity']

    def __init__(self, *args, **kwargs):
        pass


class TestDeviceModel(IsolatedAsyncioTestCase):
# ###############################################

    async def asyncSetUp(self):
        uri = 'sqlite+aiosqlite://'
        await db.init_db(uri)

    async def asyncTearDown(self):
        pass

    async def test_create(self):
        obj = Device(
            interface='__main__.MyClass',
            attributes={'address': '11:22:33:44:55:66'},
            name='11-22-33-44-55-66',
        )

        await obj.insert()
        assert obj.id >= 0
        assert obj.attributes['address'] == '11:22:33:44:55:66'
        assert 'vpd' in obj.as_dict()['reports']


if __name__ == '__main__':
    unittest.main()



from sqlalchemy.sql import text
from potnanny.models.device import Device
from potnanny.models.measurement import Measurement
from potnanny.models.interface import execute_statement


async def get_device_mtypes(pk):
    stmt = text("""
        SELECT
            m.type
        FROM
            measurements as m
        WHERE
            m.device_id = %d
        GROUP BY
            m.type
        ORDER BY
            m.type ASC
    """ % pk)

    try:
        rows = await execute_statement(stmt)
        return [r[0] for r in rows.all()]
    except:
        return []



from aiohttp import web

routes = web.RouteTableDef()

@routes.get('/api/v1.0/tests/200')
async def get_200(request):
    return web.json_response({
        "status": "ok", "msg": "test 200 ok"}, status=200)

@routes.get('/api/v1.0/tests/401')
async def get_401(request):
    return web.json_response({
        "status": "error", "msg": "test not authorized"}, status=401)


@routes.get('/api/v1.0/tests/403')
async def get_403(request):
    return web.json_response({
        "status": "error", "msg": "test forbidden"}, status=403)


@routes.get('/api/v1.0/tests/404')
async def get_403(request):
    return web.json_response({
        "status": "error", "msg": "test not found"}, status=404)



import logging
from aiohttp import web
from aiohttp_session import get_session, new_session
from potnanny.utils import verify_password, hash_password
from potnanny.models.user import User
from potnanny.models.keychain import Keychain
from potnanny.models.interface import ObjectInterface
from .decorators import login_required

routes = web.RouteTableDef()
logger = logging.getLogger(__name__)

@routes.post('/api/v1.0/login')
async def login(request):
    data = await request.json()
    user = await ObjectInterface(User).get_by_name(data['username'])

    if not user or not verify_password(data['password'], user.password):
        return web.json_response({
            "status": "error", "msg": "login failed"}, status=401)

    session = await new_session(request)
    session['username'] = user.name
    session['roles'] = user.roles
    session.changed()

    resp = web.json_response({
        "status": "ok", "msg": "authenticated ok"}, status=200)

    # for optional features
    try:
        kc = await ObjectInterface(Keychain).get_by_name('features')
        opts = {
            'username': user.name,
            'roles': user.roles,
            'room_limit': kc.attributes['room_limit'],
            'device_limit': kc.attributes['device_limit'] }

        resp.set_cookie("POTNANNY", opts, samesite="None", secure=True)
    except Exception as x:
        logger.warning(x)
        pass

    return resp


@routes.post('/api/v1.0/logout')
@login_required
async def logout(request):
    session = await get_session(request)
    session.invalidate()

    return web.json_response({
        "status": "ok", "msg": "logged out"}, status=200)


@routes.post('/api/v1.0/pwreset')
@login_required
async def pwreset(request):
    """
    reset password. requires data:
        username:
        current: (current password for user)
        password: (new password for user)
    """

    data = await request.json()
    logger.debug(data)

    # missing required data?
    for k in ['username','password','current']:
        if k not in data:
            return web.json_response({
                "status": "error", "msg": f"field {k} must be provided"},
                status=400)

    # user exists?
    user = await ObjectInterface(User).get_by_name(data['username'])
    if not user:
        return web.json_response({
            "status": "error", "msg": f"username {data['username']} not found"},
            status=404)

    # invalid current password?
    match = verify_password(data['current'], user.password)
    if not match:
        return web.json_response({
            "status": "error", "msg": f"bad password"},
            status=403)

    # hash the password and update the user
    opts = {'password': hash_password(data['password'])}
    obj = await ObjectInterface(User).update(user.id, opts)
    return web.json_response({
        "status": "ok", "msg": "password reset successfully"}, status=200)



import os
import scrypt
import hashlib
from potnanny.utils.serial import SERIAL_NUMBER

MYSALT = os.getenv('POTNANNY_SECRET', SERIAL_NUMBER)

def hash_password(password):
    secret = scrypt.hash(password, MYSALT)
    return hashlib.sha256(secret).hexdigest()

def verify_password(guessed, hashed):
    secret = hash_password(guessed)
    return secret == hashed



import logging
from aiohttp import web
from marshmallow import Schema, fields, EXCLUDE
from potnanny.models.control import Control
from potnanny.models.interface import ObjectInterface
from .decorators import login_required

routes = web.RouteTableDef()
logger = logging.getLogger(__name__)

class ControlSchema(Schema):
    class Meta:
        unknown = EXCLUDE

    name = fields.String()
    device_id = fields.Integer()
    outlet = fields.Integer()
    attributes = fields.Dict(allow_none=True)


@routes.get('/api/v1.0/controls')
@login_required
async def get_list(request):
    params = request.rel_url.query
    objects = await ObjectInterface(Control).get_all()
    if not objects:
        return web.json_response({
            "status": "error", "msg": "no data"}, status=404)

    payload = [o.as_dict() for o in objects]
    return web.json_response({
        "status": "ok", "msg": payload}, status=200)


@routes.post('/api/v1.0/controls')
@login_required
async def create(request):
    jsondata = await request.json()
    schema = ControlSchema()
    data = schema.load(jsondata)

    try:
        obj = Control(**data)
        await obj.insert()
        return web.json_response({
            "status": "ok", "msg": obj.as_dict()}, status=201)
    except Exception as x:
        return web.json_response({
            "status": "error", "msg": str(x) }, status=500)


@routes.get('/api/v1.0/controls/{pk}')
@login_required
async def get(request):
    payload = None
    pk = int(request.match_info['pk'])
    obj = await ObjectInterface(Control).get_by_id(pk)
    if not obj:
        return web.json_response({
            "status": "error", "msg": f"object id {pk} not found"},
            status=404)

    return web.json_response({
        "status": "ok", "msg": obj.as_dict()}, status=200)


@routes.patch('/api/v1.0/controls/{pk}')
@login_required
async def patch(request):
    pk = int(request.match_info['pk'])
    jsondata = await request.json()
    schema = ControlSchema()
    data = schema.load(jsondata)
    obj = await ObjectInterface(Control).update(pk, data)
    if not obj:
        return web.json_response({
            "status": "error", "msg": f"object id {pk} not found"},
            status=404)

    return web.json_response({
        "status": "ok", "msg": obj.as_dict()}, status=200)


@routes.delete('/api/v1.0/controls/{pk}')
@login_required
async def delete(request):
    pk = int(request.match_info['pk'])
    await ObjectInterface(Control).delete(pk)
    return web.json_response({
        "status": "ok", "msg": ""}, status=200)



import asyncio
import logging
from bleak import BleakScanner
from potnanny.locks import LOCKS

logger = logging.getLogger(__name__)


class BLEScanner:

    def __init__(self, *args, **kwargs):
        self.lock = LOCKS['bluetooth']
        for k, v in kwargs.items():
            if hasattr(self, k):
                setattr(self, k, v)


    async def scan_advertised(self, seconds=3, callback=None):
        """
        scan for BLE device advertisement data

        args:
            - seconds (int)
            - discovery callback function (optional)
        returns:
            - list
        """

        discovered = []

        def default_callback(reporting, advertisement):
            nonlocal discovered
            discovered.append((reporting, advertisement))

        async def callback_scan(cb, timeout=1):
            try:
                scanner = BleakScanner()
                scanner.register_detection_callback(cb)
                await scanner.start()
                await asyncio.sleep(timeout)
                await scanner.stop()
            except Exception as x:
                logger.warning(x)

        if callback is None:
            callback = default_callback

        if self.lock:
            async with self.lock:
                await callback_scan(callback, seconds)
        else:
            await callback_scan(callback, seconds)

        return discovered


    async def scan_advertised_filter(self, addresses, seconds=3, callback=None):
        """
        Returns list of advertised data, only for devices matching a list of
        MAC addresses
        """

        filtered = []
        results = await self.scan_advertised(seconds, callback)
        if not addresses:
            return results

        for d in results:
            if d[0].address in addresses:
                filtered.append(d)

        return filtered


    async def scan_devices(self):
        """
        Scan only for devices reporting
        """

        devices = None
        if self.lock:
            async with self.lock:
                devices = await BleakScanner.discover()
        else:
            devices = await BleakScanner.discover()

        return devices




import asyncio
import logging
import signal
import ssl
import potnanny.database as db
from aiohttp import web
from potnanny.config import Config
from potnanny.api import init_api
from potnanny.events import STOP_EVENT
from potnanny.controllers.worker import run_worker, stop_worker
from potnanny.locks import init_locks
from potnanny.utils import resolve_path, load_serial_number


logger = logging.getLogger(__name__)


def handle_kill(*args):
    """
    Gracefully allow asyncio processes to quit
    """

    STOP_EVENT.set()


async def init_app(config=Config()):
    """
    Initialize and run the app
    """

    # handle signal kill properly
    signal.signal(signal.SIGTERM, handle_kill)

    # load sn into global
    await load_serial_number()

    # init db and tables
    await db.init_db(config.database_uri)

    await init_locks()
    app = init_api()

    # add stop/start for the worker
    app.on_startup.append(on_startup)
    app.on_cleanup.append(on_cleanup)

    runner = web.AppRunner(app)
    await runner.setup()

    # we run the webserver on 8080, and let nginx proxy handle the https/ssl
    site = web.TCPSite(runner, '0.0.0.0', port=8080)
    await site.start()

    while not STOP_EVENT.is_set():
        await asyncio.sleep(1)

    logger.debug("Exiting")
    await runner.cleanup()


async def on_startup(app):
    await run_worker()


async def on_cleanup(app):
    await stop_worker()



import copy
import logging
from potnanny.utils import (calculate_vpd, convert_to_fahrenheit,
    flatten_list, utcnow)

logger = logging.getLogger(__name__)


class Parser:
    """
    Raw measurement parser and formatter
    """

    def __init__(self, *args, **kwargs):
        self.convert_c = False
        self.leaf_offset = -2
        self.now = utcnow()

        for k,v in kwargs.items():
            if hasattr(self, k):
                setattr(self, k, v)


    def parse(self, data: list) -> list:
        """
        Parse raw bulk measurement data from all devices in to a list of
        individual formatted measurements.

        args:
            - list
        returns:
            - list
        """

        results = []
        clean = flatten_list(data)
        for c in clean:
            results += self._extract(c)

        for m in results:
            m['created'] = self.now

        return results


    def _extract(self, data: dict) -> list:
        """
        Extract list of individual measurements from device polled data

        args:
            - dict (meas data from a device)
        returns:
            - list [
                {device_id: 1, type: temperature, value: 24.1},
                {device_id: 1, type: humidity, value: 55.1},
            ]
        """

        measurements = []
        if data is None:
            logger.debug("received None instead of a dict?")
            return []

        try:
            data['device_id'] = data.pop('id')
            values = data.pop('values')

            # vpd calcs are done first
            if 'temperature' in values and 'humidity' in values:
                try:
                    m = copy.copy(data)
                    t = values['temperature'] + self.leaf_offset
                    h = values['humidity']
                    m.update({
                        'type': 'vpd',
                        'value': calculate_vpd(t, h) })
                    measurements.append(m)
                except Exception as x:
                    logger.warning(x)

            for key, value in values.items():
                m = copy.copy(data)

                if key == 'temperature' and self.convert_c is True:
                    value = convert_to_fahrenheit(value)

                m.update({'type': key, 'value': value, 'created': self.now})
                measurements.append(m)

        except Exception as x:
            logger.warning(x)

        return measurements



import logging
from aiohttp import web
from potnanny.controllers.environment import (get_environments,
    get_room_environments)
from .decorators import login_required

routes = web.RouteTableDef()
logger = logging.getLogger(__name__)

@routes.get('/api/v1.0/environments')
@login_required
async def all_environments(request):
    results = await get_environments()
    if not results:
        return web.json_response({
            "status": "error", "msg": "no data"}, status=404)

    return web.json_response({
        "status": "ok", "msg": results}, status=200)


@routes.get('/api/v1.0/environments/{pk}')
@login_required
async def room_environment(request):
    pk = int(request.match_info['pk'])
    results = await get_room_environments(pk)
    if not results:
        return web.json_response({
            "status": "error", "msg": "no data"}, status=404)

    return web.json_response({
        "status": "ok", "msg": results}, status=200)



import asyncio
import unittest
import datetime
import json
import potnanny.database as db
from unittest import IsolatedAsyncioTestCase
from potnanny.models import Room
from potnanny.models.utils import (get_class_objects, update_class_object,
    delete_class_object)


class TestUtilsModel(IsolatedAsyncioTestCase):
# ###############################################

    async def asyncSetUp(self):
        uri = 'sqlite+aiosqlite://'
        await db.init_db(uri)

    async def asyncTearDown(self):
        pass

    async def test_create(self):
        for i in (1,2,3):
            r = Room(name='test room %d' % i)
            await r.insert()

        results = await get_class_objects(Room)
        assert len(results) == 3

        room = await get_class_objects(Room, 1)
        assert type(room) is Room

        missing = await get_class_objects(Room, 99)
        assert missing is None

if __name__ == '__main__':
    unittest.main()




class WeekdayMap:
    """A utility Class which maps days of week to a numeric value. Used for scheduling.

    Usage:
        >>> WeekdayMap.weekdays_from_value(7)
        ['Thursday', 'Friday', 'Saturday']

        >>> WeekdayMap.DAYS
        {
            64: 'Sunday',
            32: 'Monday',
            16: 'Tuesday',
            8:  'Wednesday',
            4:  'Thursday',
            2:  'Friday',
            1:  'Saturday',
        }
    """

    DAYS = {
        64: 'Sunday',
        32: 'Monday',
        16: 'Tuesday',
        8:  'Wednesday',
        4:  'Thursday',
        2:  'Friday',
        1:  'Saturday',
    }

    @classmethod
    def weekdays_from_value(cls, val):
        """Get list of weekday names that are valid for particular run value.

        args:
          - VALUE (int) 0-127
        returns:
            a list
        """
        if val < 0 or val > 127:
            raise ValueError("Value must be from 0 to 127")

        results = []
        for n in sorted(cls.DAYS.keys(), reverse=True):
            if (val & n):
                dow = cls.DAYS[n]
                results.append(dow)

        return results




import asyncio
import logging
from marshmallow import Schema, fields, EXCLUDE, INCLUDE
from sqlalchemy import (Column, Integer, DateTime, Unicode, UnicodeText,
    ForeignKey, func)
from sqlalchemy.orm import relationship
from potnanny.database import Base
from potnanny.models.mixins import CRUDMixin


logger = logging.getLogger(__name__)


class RoomSchema(Schema):
    class Meta:
        unknown = EXCLUDE

    name = fields.String()
    notes = fields.String(allow_none=True)


class Room(Base, CRUDMixin):
    __tablename__ = 'rooms'

    id = Column(Integer, primary_key=True)
    name = Column(Unicode(128), nullable=False, unique=True)
    notes = Column(UnicodeText, nullable=True, unique=False)
    created = Column(DateTime, server_default=func.now())

    # make relationships compatible with asyncio
    __mapper_args__ = {"eager_defaults": True}

    # relationships
    devices = relationship('Device',
        backref='room',
        cascade='all, delete',
        lazy='subquery')


    def __repr__(self):
        return "<Room(id={}, name={})>".format(self.id, self.name)


    def as_dict(self, explore=True):
        """
        present object as a dict.
        args:
            - explore: (boolean) recursively explore relationships and include
              their data as well
        returns:
            dict
        """

        data = {
            'id': self.id,
            'name': self.name,
            'notes': self.notes,
            'created': self.created.isoformat() + "Z",
        }
        if explore:
            try:
                data['devices'] = [i.as_dict() for i in self.devices]
            except:
                pass

        return data



import logging
from aiohttp import web
from potnanny.models.device import Device, DeviceSchema
from potnanny.models.interface import ObjectInterface
from potnanny.controllers.device import get_device_mtypes
from potnanny.controllers.outlet import switch_device_outlet
from potnanny.controllers.worker import Worker
from potnanny.controllers.graph import device_graph
from potnanny.locks import LOCKS
from .decorators import login_required

routes = web.RouteTableDef()
logger = logging.getLogger(__name__)

@routes.get('/api/v1.0/devices')
@login_required
async def get_list(request):
    objects = await ObjectInterface(Device).get_all()
    if not objects:
        return web.json_response({
            "status": "error", "msg": "object not found"},
            status=404)

    return web.json_response({
        "status": "ok", "msg": [o.as_dict()  for o in objects] }, status=200)


@routes.post('/api/v1.0/devices')
@login_required
async def create(request):
    jsondata = await request.json()
    schema = DeviceSchema()
    data = schema.load(jsondata)
    try:
        obj = Device(**data)
        await obj.insert()
        return web.json_response({
            "status": "ok", "msg": obj.as_dict()}, status=201)
    except Exception as x:
        return web.json_response({
            "status": "error", "msg": str(x) }, status=500)


@routes.get('/api/v1.0/devices/{pk}')
@login_required
async def get(request):
    pk = int(request.match_info['pk'])
    obj = await ObjectInterface(Device).get_by_id(pk)
    if not obj:
        return web.json_response({
            "status": "error", "msg": f"object with id {pk} not found"},
            status=404)

    return web.json_response({
        "status": "ok", "msg": obj.as_dict()}, status=200)


@routes.patch('/api/v1.0/devices/{pk}')
@login_required
async def patch(request):
    changed = 0
    pk = int(request.match_info['pk'])
    jsondata = await request.json()
    schema = DeviceSchema()
    data = schema.load(jsondata)

    obj = await ObjectInterface(Device).update(pk, data)
    if not obj:
        return web.json_response({
            "status": "error", "msg": f"object with id {pk} not found"},
            status=404)

    return web.json_response({
        "status": "ok", "msg": obj.as_dict()}, status=200)


@routes.delete('/api/v1.0/devices/{pk}')
@login_required
async def delete(request):
    pk = int(request.match_info['pk'])
    await ObjectInterface(Device).delete(pk)
    return web.json_response({
        "status": "ok", "msg": ""}, status=200)


@routes.get('/api/v1.0/devices/{pk}/measurement-types')
@login_required
async def get_measurement_types(request):
    pk = int(request.match_info['pk'])
    results = await get_device_mtypes(pk)
    if not results:
        return web.json_response({
            "status": "error", "msg": f"device id {pk} no measurements"},
            status=404)

    return web.json_response({
        "status": "ok", "msg": results}, status=200)


@routes.post('/api/v1.0/devices/{pk}/poll')
@login_required
async def device_poll(request):
    pk = int(request.match_info['pk'])
    worker = Worker()
    await worker.poll_device(pk)
    return web.json_response({
        "status": "ok", "msg": ""}, status=200)


@routes.post('/api/v1.0/devices/{pk}/{outlet}/on')
@login_required
async def device_on(request):
    pk = int(request.match_info['pk'])
    outlet = int(request.match_info['outlet'])
    state = 1
    result = await switch_device_outlet(pk, outlet, state)
    if result is None or result != state:
        return web.json_response({
            "status": "error", "msg": f"error switching device {pk} outlet {outlet}"},
            status=500)

    return web.json_response({
        "status": "ok", "msg": {'state': state}}, status=200)


@routes.post('/api/v1.0/devices/{pk}/{outlet}/off')
@login_required
async def device_off(request):
    pk = int(request.match_info['pk'])
    outlet = int(request.match_info['outlet'])
    state = 0
    result = await switch_device_outlet(pk, outlet, state)
    if result is None or result != state:
        return web.json_response({
            "status": "error", "msg": f"error switching device {pk} outlet {outlet}"},
            status=500)

    return web.json_response({
        "status": "ok", "msg": {'state': state}}, status=200)


@routes.get('/api/v1.0/devices/{pk}/keycode')
@login_required
async def get_keycode(request):
    pk = int(request.match_info['pk'])
    device = await ObjectInterface(Device).get_by_id(pk)
    if not device:
        return web.json_response({
            "status": "error", "msg": "not found"}, status=404)

    if (not hasattr(device.plugin, 'is_key_required') or
        device.plugin.is_key_required is not True):
        return web.json_response({
            "status": "error", "msg": "not a valid request for this device"},
                status=405)

    secret = None
    try:
        if 'bluetooth' in LOCKS and LOCKS['bluetooth'] is not None:
            async with LOCKS['bluetooth'] as lock:
                secret = await device.plugin.scan_key()
        else:
            secret = await device.plugin.scan_key()
    except Exception as x:
        logger.debug(x)

    if secret:
        return web.json_response({
            "status": "ok", "msg": secret}, status=200)
    else:
        return web.json_response({
            "status": "error", "msg": "key code not detected before timeout"},
            status=404)


@routes.get('/api/v1.0/devices/{pk}/graph')
@login_required
async def graph(request):
    pk = int(request.match_info['pk'])
    mtype = request.rel_url.query.get('measurement', None)
    hours = int(request.rel_url.query.get('hours', 8))

    mygraph = await device_graph(pk, hours, mtype)
    if not mygraph:
        return web.json_response({
            "status": "error", "msg": "no data"},
            status=404)

    return web.json_response({
        "status": "ok", "msg": mygraph}, status=200)



import re

def evaluate(equation: str) -> bool:
    """
    evaluate a conditional equation str, like:
    ("22 >= 80")        # False
    ("11.1 == 13.9")    # False
    ("2 eq 2.0")        # True

    args:
        - a string statement, like "22 lt 100"
    returns:
        bool (True|False)
    """

    atoms = re.split(r'\s+', equation)
    if len(atoms) != 3:
        raise ValueError("Poorly formed equation '%s'" % equation)

    a = float(atoms[0])
    b = float(atoms[2])
    oper = atoms[1]

    if oper in ['lt','<'] and a < b:
        return True
    elif oper in ['le','<='] and a <= b:
        return True
    elif oper in ['gt','>'] and a > b:
        return True
    elif oper in ['ge','>='] and a >= b:
        return True
    elif oper in ['eq','==', '='] and a == b:
        return True
    elif oper in ['ne','!='] and a != b:
        return True

    return False



import asyncio
import unittest
import datetime
from unittest import IsolatedAsyncioTestCase
from potnanny.locks import LOCKS, init_locks


class TestLockFuncs(IsolatedAsyncioTestCase):
# ###############################################

    async def test_lock(self):
        await init_locks()
        assert LOCKS['bluetooth'] is not None
        assert LOCKS['db'] is not None
        assert LOCKS['gpio'] is not None

    async def test_lock_types(self):
        await init_locks()
        assert type(LOCKS['bluetooth']) is asyncio.Lock
        assert type(LOCKS['db']) is asyncio.Lock
        assert type(LOCKS['gpio']) is asyncio.Lock


if __name__ == '__main__':
    unittest.main()



import asyncio
import logging
import random
from potnanny.utils.scanner import BLEScanner


logger = logging.getLogger(__name__)


class BLEInterface(BLEScanner):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # the reference to self.lock is inherited from BLEScanner
        for k, v in kwargs.items():
            if hasattr(self, k):
                setattr(self, k, v)


    async def poll_devices(self, devices: list) -> list:
        """
        poll measurements from list of devices

        args:
            - list of Device objects
        returns:
            - list of measurements
        """

        results = []

        # divide into lists of Readers or Poller devices
        readers = self._reader_list(devices)
        clients = self._poller_list(devices)

        # run polling and gather results when done
        task_list = []
        if readers:
            task_list.append(self._poll_readers(readers))

        if clients:
            task_list.append(self._poll_clients(clients))

        results = await asyncio.gather(*task_list)
        return results


    def _reader_list(self, devices):
        # parse list of devices and return only ble advertisement readers
        results = []
        for d in devices:
            if (not hasattr(d.plugin, 'is_bluetooth') or
                d.plugin.is_bluetooth is not True):
                continue

            if (not hasattr(d.plugin, 'is_reader') or
                d.plugin.is_reader is not True):
                continue

            results.append(d)

        logger.debug("Reader List: %s" % results)
        return results


    def _poller_list(self, devices):
        # parse list of devices and return only ble pollable clients
        results = []
        for d in devices:
            if (not hasattr(d.plugin, 'is_bluetooth') or
                d.plugin.is_bluetooth is not True):
                continue

            if (not hasattr(d.plugin, 'is_pollable') or
                d.plugin.is_pollable is not True):
                continue

            results.append(d)

        logger.debug("Poller List: %s" % results)
        return results


    async def _poll_clients(self, devices):
        """
        poll devices that we connect to as a client

        args:
            - list of devices to collect for
        """

        results = []
        for d in devices:
            await asyncio.sleep(random.random())
            try:
                data = None
                if self.lock:
                    async with self.lock:
                        data = await d.poll()
                else:
                    data = await d.poll()

                if data is not None:
                    results.append(data)

            except Exception as x:
                logger.warning(x)

        return results


    async def _poll_readers(self, devices):
        """
        poll devices that Read Advertised Data(RAD) from bluetooth

        args:
            - list of devices to collect for
            - timestamp for collected measurements
        """

        results = []
        found = {d.id:None for d in devices}
        tries = 10
        success = False
        while tries and not success:
            addresses = [d.attributes['address'] for d in devices if (found[d.id] is None)]
            scans = await self.scan_advertised_filter(addresses)
            found.update(self._parse_advertised_data(devices, scans))

            if None not in found.values():
                success = True

            if not success:
                tries -= 1
                await asyncio.sleep(random.random())

        results += found.values()
        return results


    def _parse_advertised_data(self, devices, data):
        """
        parse advertisement results

        args:
            - list of devices
            - list of data points
        returns:
            dict of {
                device_id: measurement,
            }
        """

        results = {}
        for bundle in data:
            if bundle is None:
                continue
            reporting, advertisement = bundle
            fingerprint = {'address': reporting.address, 'name': reporting.name}
            for d in devices:
                if d.room_id is None:
                    continue

                if not d.plugin.__class__.recognize_this(fingerprint):
                    continue

                try:
                    if d.attributes['address'].upper() != reporting.address.upper():
                        continue
                except Exception as x:
                    logger.warning(x)
                    continue

                try:
                    data = d.read_advertisement(reporting, advertisement)
                    if data:
                        results[d.id] = data
                    else:
                        logger.warning("No data: Device:%d, Reporting:%s, Adv:%s" % (
                            d.id, reporting, advertisement))
                except Exception as x:
                    logger.warning(x)

        return results



import os
import unittest
import datetime
from potnanny.utils.pids import (pid_to_file, pid_from_file, check_pid, 
    is_running)


class TestPidFuncs(unittest.TestCase):
# ###############################################

    def test_running(self):
        path = '/tmp/test.pid'
        pid = os.getpid()
        pid_to_file(pid, path)
        assert is_running(path) == pid
        os.unlink(path)

    def test_pid_file(self):
        value = 9999
        path = '/tmp/test.pid'
        pid_to_file(value, path)

        assert os.path.exists(path)
        assert pid_from_file(path) == value

        os.unlink(path)

    def test_check_pid(self):
        pid = os.getpid()
        assert check_pid(pid) is True


if __name__ == '__main__':
    unittest.main()



import asyncio
import unittest
from unittest import IsolatedAsyncioTestCase
from potnanny.utils.scanner import BLEScanner


class TestBleInterface(IsolatedAsyncioTestCase):
# ###############################################

    async def test_scan_devices(self):
        obj = BLEScanner()
        devices = await obj.scan_devices()
        assert devices

    async def test_scan_advertised(self):
        obj = BLEScanner()
        results = await obj.scan_advertised()
        assert results


if __name__ == '__main__':
    unittest.main()



import logging
from sqlalchemy import (Column, Integer, DateTime, Unicode, UnicodeText,
    ForeignKey, Boolean, func)
from potnanny.database import Base
from potnanny.models.mixins import CRUDMixin


logger = logging.getLogger(__name__)


class Error(Base, CRUDMixin):
    __tablename__ = 'errors'

    id = Column(Integer, primary_key=True)
    title = Column(Unicode(64), nullable=False, unique=False)
    message = Column(UnicodeText, nullable=False, unique=False)
    severity = Column(Integer, nullable=False, unique=False, default=1)
    acknowledged = Column(Boolean, nullable=False, unique=False, default=False)
    created = Column(DateTime, server_default=func.now())

    # severity = 1: minor, 2: major, 3: critical

    def __repr__(self):
        return "<Error(id={}, title={})>".format(self.id, self.title)


    def as_dict(self):
        """
        present object as a dict.
        args:
        returns:
            dict
        """

        data = {
            'id': self.id,
            'title': self.title,
            'message': self.message,
            'severity': self.severity,
            'acknowledged': self.acknowledged,
            'created': self.created.isoformat() + "Z",
        }

        return data



import logging
from marshmallow import Schema, fields, EXCLUDE, INCLUDE
from sqlalchemy import (Column, Integer, Unicode, UnicodeText, Float, DateTime,
    ForeignKey, func)
from potnanny.database import Base
from potnanny.models.mixins import CRUDMixin


logger = logging.getLogger(__name__)


class MeasurementSchema(Schema):
    class Meta:
        unknown = EXCLUDE

    type = fields.String(allow_none=False)
    value = fields.Float(allow_none=False)
    created = fields.Raw(allow_none=True)
    device_id = fields.Integer(allow_none=False)


class Measurement(Base, CRUDMixin):
    __tablename__ = 'measurements'

    id = Column(Integer, primary_key=True)
    type = Column(Unicode(24), nullable=False, index=True)
    value = Column(Float, nullable=False)
    created = Column(DateTime, server_default=func.now())

    # make relationships compatible with asyncio
    __mapper_args__ = {"eager_defaults": True}

    # relationships
    device_id = Column(Integer, ForeignKey('devices.id'), nullable=True)

    def as_dict(self):
        return {
            'id': self.id,
            'type': self.type,
            'value': self.value,
            'created': self.created.isoformat() + "Z",
            'device_id': self.device_id,
        }



import asyncio
import logging
import datetime
import potnanny.database as db
from potnanny.locks import LOCKS
from sqlalchemy.inspection import inspect

logger = logging.getLogger(__name__)

class CRUDMixin(object):
    """
    Adds async CRUD functionality to our base sqlalchemy models
    """

    async def insert(self):
        """insert self into db"""

        async def perform():
            async with db.session() as session:
                session.add(self)
                await session.commit()

        await self._execute(perform)


    async def delete(self):
        """delete self from db"""

        async def perform():
            async with db.session() as session:
                await session.delete(self)
                await session.commit()

        await self._execute(perform)


    async def update(self):
        """update self from db"""

        if hasattr(self, 'modified'):
            self.modified = datetime.datetime.utcnow()

        async def perform():
            async with db.session() as session:
                await session.commit()

        await self._execute(perform)


    async def _execute(self, f):
        """execute the session function"""

        if 'db' in LOCKS and LOCKS['db'] is not None:
            async with LOCKS['db']:
                await f()
        else:
            await f()



import asyncio

async def run(cmd: str):
    """
    Async run shell command
    args:
        command string
    returns:
        tuple (exit-code, stdout, stderr)
    """

    proc = await asyncio.create_subprocess_shell(
        cmd,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE)

    stdout, stderr = await proc.communicate()
    return (proc.returncode, stdout, stderr)



import logging
from aiohttp import web
from potnanny.utils import iso_from_sqlite
from potnanny.models.room import Room, RoomSchema
from potnanny.models.interface import ObjectInterface
from potnanny.controllers.graph import room_graph
from .decorators import login_required


routes = web.RouteTableDef()
logger = logging.getLogger(__name__)


@routes.get('/api/v1.0/rooms')
@login_required
async def get_list(request):
    objects = await ObjectInterface(Room).get_all()
    if not objects:
        return web.json_response({
            "status": "error", "msg": "no data"}, status=404)

    payload = [o.as_dict() for o in objects]

    return web.json_response({
        "status": "ok", "msg": payload}, status=200)


@routes.post('/api/v1.0/rooms')
@login_required
async def create(request):
    jsondata = await request.json()
    schema = RoomSchema()
    data = schema.load(jsondata)

    try:
        obj = Room(**data)
        await obj.insert()
        return web.json_response({
            "status": "ok", "msg": obj.as_dict()}, status=201)
    except Exception as x:
        return web.json_response({
            "status": "error", "msg": str(x) }, status=500)


@routes.get('/api/v1.0/rooms/{pk}')
@login_required
async def get(request):
    payload = None
    pk = int(request.match_info['pk'])
    obj = await ObjectInterface(Room).get_by_id(pk)
    if not obj:
        return web.json_response({
            "status": "error", "msg": f"object id {pk} not found"},
            status=404)

    return web.json_response({
        "status": "ok", "msg": obj.as_dict()}, status=200)


@routes.patch('/api/v1.0/rooms/{pk}')
@login_required
async def patch(request):
    pk = int(request.match_info['pk'])
    jsondata = await request.json()
    schema = RoomSchema()
    data = schema.load(jsondata)

    obj = await ObjectInterface(Room).update(pk, data)
    if not obj:
        return web.json_response({
            "status": "error", "msg": f"object with id {pk} not found"},
            status=404)

    return web.json_response({
        "status": "ok", "msg": obj.as_dict()}, status=200)


@routes.delete('/api/v1.0/rooms/{pk}')
@login_required
async def delete(request):
    pk = int(request.match_info['pk'])
    await ObjectInterface(Room).delete(pk)
    return web.json_response({
        "status": "ok", "msg": ""}, status=202)


@routes.get('/api/v1.0/rooms/{pk}/graph')
@login_required
async def graph(request):
    pk = int(request.match_info['pk'])
    mtype = request.rel_url.query.get('measurement', None)
    hours = int(request.rel_url.query.get('hours', 8))

    graph = await room_graph(pk, mtype, hours)
    if not graph:
        return web.json_response({
            "status": "error", "msg": "no data"},
            status=404)

    return web.json_response({
        "status": "ok", "msg": graph}, status=200)



from .times import (utcnow, datetime_from_iso, datetime_from_sqlite,
    iso_from_sqlite)
from .numbers import is_number
from .paths import resolve_path
from .temperature import convert_to_fahrenheit
from .vpd import calculate_vpd
from .network import has_www, aio_has_www
from .lists import flatten_list
from .pids import PIDFILE, is_running, pid_from_file, pid_to_file, check_pid
from .plugins import load_plugins
from .shell import run
from .pw import hash_password, verify_password
from .eval import evaluate
from .serial import load_serial_number

__all__ = [
    utcnow,
    datetime_from_iso,
    datetime_from_sqlite,
    iso_from_sqlite,
    resolve_path,
    convert_to_fahrenheit,
    calculate_vpd,
    is_number,
    has_www,
    aio_has_www,
    flatten_list,
    PIDFILE, is_running, pid_from_file, pid_to_file, check_pid,
    load_plugins,
    run,
    hash_password, verify_password,
    evaluate,
    load_serial_number
]



import os
import unittest
from potnanny.utils.numbers import is_number


class TestUtilFuncs(unittest.TestCase):
# ###############################################

    def test_int(self):
        n = "13"
        assert is_number(n) == int

    def test_float(self):
        n = "13.0"
        assert is_number(n) == float

    def test_bogus(self):
        n = "bogus"
        assert is_number(n) == None


if __name__ == '__main__':
    unittest.main()



import re
import asyncio
import logging
import potnanny.database as db
from potnanny.models.interface import ObjectInterface
from potnanny.plugins.base import BluetoothDevicePlugin
from potnanny.models.device import Device
from potnanny.utils.scanner import BLEScanner
from potnanny.locks import LOCKS

logger = logging.getLogger(__name__)


class DiscoveryClient:
    def __init__(self, *args, **kwargs):
        self.scanner = BLEScanner()


    async def discover(self):
        results = []
        existing = await ObjectInterface(Device).get_all()
        found = await self.scanner.scan_devices()
        devices = self.recognized_devices([f for f in found])
        for entry in devices:
            device, interface, attrs = entry
            if self.is_existing_device(existing, device, interface):
                continue

            opts = {
                'name': device.name,
                'interface': interface,
                'attributes': {'address': device.address},
                'room_id': None }

            # manufacturer bluetooth advertised names are often stupid, and
            # meaningless to the end user.
            # here, we replace it with the plugin name attribute to be more
            # helpful.
            try:
                klass = BluetoothDevicePlugin.get_named_class(opts['interface'])
                if klass:
                    opts['name'] = klass.name
            except:
                pass

            try:
                opts['attributes'].update(attrs)
            except:
                pass

            results.append(opts)

        for d in results:
            try:
                obj = Device(**d)
                await obj.insert()
            except Exception as x:
                logger.warning(x)

        return results


    async def insert_devices(self, devices):
        async def db_execute(f):
            """execute the session function"""

            if 'db' in LOCKS and LOCKS['db'] is not None:
                async with LOCKS['db']:
                    await f()
            else:
                await f()

        async def perform():
            try:
                async with db.session() as session:
                    for d in devices:
                        obj = Device(**d)
                        session.add(obj)

                    await session.commit()
            except Exception as x:
                logger.warning(x)

        await db_execute(perform)


    def recognized_devices(self, devices):
        """
        Return a filtered list of ble devices that we can communicate with.
        [(device, interface_name), ]
        """

        results = []
        for d in devices:
            fingerprint = {'address': d.address, 'name': d.name}
            for p in BluetoothDevicePlugin.plugins:
                attrs = None
                try:
                    if p.recognize_this(fingerprint):
                        found = '.'.join((p.__module__, p.__name__))
                        try:
                            attrs = p.attributes
                        except:
                            pass
                        results.append((d, found, attrs))
                except:
                    continue

        return results


    def is_existing_device(self, existing, device, interface):
        """
        Compare ble-device to list of existing devices, and see if match
        exists already

        args:
            - list of existing devices
            - device to inspect
            - plugin interface
        returns:
            - bool
        """

        def match_exists(e, plugin, address):
            try:
                if e.interface != plugin:
                    return False

                a1 = address.upper()
                a2 = e.attributes['address'].upper()
                if a1 == a2:
                    return True
            except Exception as x:
                logger.warning(str(x))

            return False

        for e in existing:
            if match_exists(e, interface, device.address) == True:
                return True

        return False


async def discover_new_devices():
    client = DiscoveryClient()
    await client.discover()



import asyncio
import logging
import datetime
from potnanny.utils import utcnow
from potnanny.models.schedule import Schedule
from potnanny.models.interface import ObjectInterface

logger = logging.getLogger(__name__)


async def run_schedules(now=None):
    """
    Run scheduled actions
    """

    tasks = []
    if now is None:
        now = datetime.datetime.now().replace(second=0, microsecond=0)

    schedules = await ObjectInterface(Schedule).get_all()
    for s in schedules:
        try:
            if s.runs_now() is True:
                tasks.append(s.run(now))
        except Exception as x:
            logger.warning(str(x))

    if tasks:
        results = await asyncio.gather(*tasks)



import datetime
import logging
import asyncio
from potnanny.utils import utcnow, load_plugins
from potnanny.models.device import Device
from potnanny.models.keychain import Keychain
from potnanny.models.interface import ObjectInterface
from potnanny.locks import LOCKS
from potnanny.events import WORKER_EVENT
from potnanny.controllers.poll import Poller
from potnanny.controllers.schedule import run_schedules
from potnanny.controllers.cleanup import purge_measurements


logger = logging.getLogger(__name__)
TASK = None

async def run_worker():
    """
    Convenience function, to run the worker
    """
    global TASK
    logger.debug("Starting worker task")
    worker = Worker()
    TASK = asyncio.create_task(worker.run())


async def stop_worker():
    """
    Convenience function, to signal worker to stop
    """
    logger.debug("Stopping worker task")
    WORKER_EVENT.set()


async def restart_worker():
    """
    Convenience function, to stop and restart the worker
    """
    await stop_worker()
    while TASK in asyncio.all_tasks():
        await asyncio.sleep(1)

    await run_worker()


class Worker:
    def __init__(self, *args, **kwargs):
        self.convert_c = False
        self.leaf_offset = -2
        self.polling_interval = 10
        self.plugin_path = "~/potnanny/plugins"
        self.storage_days = 7


    async def load_settings(self):
        try:
            obj = await ObjectInterface(Keychain).get_by_name('settings')
            for k, v in obj.attributes.items():
                if hasattr(self, k):
                    setattr(self, k, v)
                elif k == 'temperature_display':
                    if v in ['f','F']:
                        self.convert_c = True
        except:
            pass


    async def load_plugins(self):
        load_plugins(self.plugin_path)


    async def run(self):
        await self.load_settings()
        await self.load_plugins()

        while not WORKER_EVENT.is_set():
            unow = datetime.datetime.utcnow().replace(microsecond=0)

            # at midnight we clean old records
            if unow.hour == 0 and unow.minute == 0 and unow.second == 0:
                start = unow - datetime.timedelta(days=self.storage_days)
                logger.debug(f"cleaning up measurements older than {start}")
                t1 = asyncio.create_task(purge_measurements(start))

            if unow.second == 0:
                # run schedules at top of every minute
                t2 = asyncio.create_task(run_schedules())

                # poll devices at intervals
                if unow.minute % self.polling_interval == 0:
                    t3 = asyncio.create_task(self.poll_devices())

            await asyncio.sleep(1)


    async def poll_devices(self):
        """
        poll active devices for information
        """

        logger.debug("polling devices")
        opts = {
            'convert_c': self.convert_c,
            'leaf_offset': self.leaf_offset}

        p = Poller(**opts)
        await p.poll()


    async def poll_device(self, pk: int):
        """
        poll device for information
        """

        logger.debug(f"polling device {pk}")
        opts = {
            'convert_c': self.convert_c,
            'leaf_offset': self.leaf_offset}

        p = Poller(**opts)
        await p.poll_id(pk)



import setuptools

with open("README.md", "r", encoding="utf8") as fh:
    long_description = fh.read()

setuptools.setup(
    name='potnanny',
    version='0.1.5',
    python_requires=">=3.9",
    packages=setuptools.find_packages(),
    include_package_data=True,
    description='Potnanny grow room automation controller.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='J Leary',
    author_email='potnanny@gmail.com',
    url='https://github.com/potnanny/application',
    install_requires=[
        'requests==2.31.0',
        'scrypt==0.8.20',
        'cryptography==41.0.1',
        'aiofiles==23.1.0',
        'aiohttp==3.8.4',
        'aiohttp-jinja2==1.5.1',
        'aiohttp-cors==0.7.0',
        'aiohttp-session==2.12.0',
        'aiosqlite==0.19.0',
        'aiosmtplib==2.0.2',
        'bleak==0.20.2',
        'gpiozero==1.6.2',
        'greenlet==2.0.2',
        'marshmallow==3.19.0',
        'pyyaml==6.0',
        'python-daemon==3.0.1',
        'sqlalchemy==2.0.16',
    ],
    package_data = {
        '': ['*.html', '*.css', '*.js', '*.woff2'],
    },
    entry_points = {
        'console_scripts': [
            'potnanny=potnanny.cli:main',
        ],
    },
)



import asyncio
import unittest
import potnanny.database as db
from unittest import IsolatedAsyncioTestCase
from potnanny.models.control import Control


class TestDeviceModel(IsolatedAsyncioTestCase):
# ###############################################

    async def asyncSetUp(self):
        uri = 'sqlite+aiosqlite://'
        await db.init_db(uri)

    async def asyncTearDown(self):
        pass

    async def test_create(self):
        obj = Control(
            name='test',
            device_id=1,
            attributes={
                'input_device_id': 2,
                'type': 'temperature',
                'on': {
                    'condition': 'gt',
                    'threshold': 60
                },
                'off': {
                    'condition': 'lt',
                    'threshold': 50
                }
            }
        )
        await obj.insert()
        assert obj.eval("1 gt 0") is True
        assert obj.eval("0 lt 1") is True
        assert obj.eval("0 ge 0") is True
        assert obj.eval("1 ne 0") is True
        assert obj.eval("1 == 1") is True
        assert obj.eval("1 != 1") is False


if __name__ == '__main__':
    unittest.main()



import asyncio
import unittest
from unittest import IsolatedAsyncioTestCase
import potnanny.database as db
from potnanny.models.interface import ObjectInterface
from potnanny.models.keychain import Keychain


class TestKeychainModel(IsolatedAsyncioTestCase):
# ###############################################

    async def asyncSetUp(self):
        uri = 'sqlite+aiosqlite://'
        await db.init_db(uri)

    async def asyncTearDown(self):
        pass

    async def test_create(self):
        opts = {
            'name': 'mail-auth',
            'attributes': {
                'user': 'test@gmail.com',
                'auth_token': '1234567890',
                'refresh_token': '1234567890',
            }
        }
        obj = Keychain(**opts)

        await obj.insert()
        assert obj.id >= 0
        assert 'user' in obj.attributes
        assert obj.attributes['user'] == 'test@gmail.com'


    async def test_get_id(self):
        obj = await ObjectInterface(Keychain).get_by_id(1)
        assert obj is not None


    async def test_get_name(self):
        obj = await ObjectInterface(Keychain).get_by_name('settings')
        assert obj is not None


    async def test_update(self):
        opts = {
            'name': 'mail-auth2',
            'attributes': {
                'user': 'test@gmail.com',
                'auth_token': '1234567890',
                'refresh_token': '1234567890',
            }
        }
        obj = Keychain(**opts)
        await obj.insert()

        obj.attributes['user'] = 'modified@gmail.com'
        await obj.update()
        assert obj.attributes['user'] == 'modified@gmail.com'


    async def test_delete(self):
        opts = {
            'name': 'mail-auth3',
            'attributes': {
                'user': 'test@gmail.com',
                'auth_token': '1234567890',
                'refresh_token': '1234567890',
            }
        }
        obj = Keychain(**opts)
        await obj.insert()
        assert obj.id > 0

        pk = obj.id
        await obj.delete()

        check = await ObjectInterface(Keychain).get_by_id(pk)
        assert check is None


if __name__ == '__main__':
    unittest.main()



import asyncio
import unittest
import potnanny.database as db
from potnanny.locks import init_locks
from unittest import IsolatedAsyncioTestCase
from potnanny.models import Measurement, Device


class TestMeasurementModel(IsolatedAsyncioTestCase):
# ###############################################

    async def asyncSetUp(self):
        uri = 'sqlite+aiosqlite://'
        await db.init_db(uri)
        await init_locks()

    async def asyncTearDown(self):
        pass

    async def test_create(self):
        m = Measurement(
            device_id=1,
            type='temperature',
            value=74.2
        )
        await m.insert()
        assert m.id >= 0
        assert type(m.as_dict()) is dict
        assert m.value == 74.2

if __name__ == '__main__':
    unittest.main()



import asyncio
import unittest
import datetime
import json
import potnanny.database as db
from unittest import IsolatedAsyncioTestCase
from potnanny.models import Room, Device
from sqlalchemy.future import select
from sqlalchemy.orm import selectinload


class TestRoomModel(IsolatedAsyncioTestCase):
# ###############################################

    async def asyncSetUp(self):
        uri = 'sqlite+aiosqlite://'
        await db.init_db(uri)

    async def asyncTearDown(self):
        pass

    async def test_create(self):
        r = Room(
            name='test room 1',
            notes='test')
        await r.insert()
        assert r.id >= 0

    async def test_relationships(self):
        r = Room(name='test room 2')
        await r.insert()

        i = Device(
            name='test input 2',
            notes='test',
            interface='plugins.input.ble.test.TestInput',
            attributes={'address': '11:22:33:44:55:66'},
            room_id=r.id)
        await i.insert()

        # each of the above insert()s was performed in a different session.
        # so, we must re-query the room, to get all objects together in one
        # session...
        room = None
        async with db.session() as session:
            stmt = select(Room).filter(
                Room.name == 'test room 2').options(
                selectinload(Room.devices)
            )

            results = await session.execute(stmt)
            room = results.scalar()

        assert('devices' in room.as_dict())
        assert(type(room.as_dict()['devices']) is list)

if __name__ == '__main__':
    unittest.main()



import re
import asyncio
import logging
import datetime
from marshmallow import Schema, fields, EXCLUDE, INCLUDE
from sqlalchemy import (Column, Integer, Text, String, Float, Boolean,
    DateTime, ForeignKey, func)
from sqlalchemy.orm import relationship
from potnanny.database import Base
from potnanny.controllers.outlet import switch_device_outlet
from potnanny.utils import evaluate
from potnanny.models.mixins import CRUDMixin
from potnanny.models.ext import MutableDict, JSONEncodedDict


logger = logging.getLogger(__name__)


class ControlSchema(Schema):
    class Meta:
        unknown = EXCLUDE

    name = fields.String()
    outlet = fields.Integer()
    device_id = fields.Integer()
    attributes = fields.Dict(allow_none=True)


class Control(Base, CRUDMixin):
    __tablename__ = 'controls'

    id = Column(Integer, primary_key=True)
    name = Column(String(128), nullable=False, unique=False)
    outlet = Column(Integer, nullable=False)
    created = Column(DateTime, server_default=func.now())
    attributes = Column(MutableDict.as_mutable(JSONEncodedDict), nullable=True)

    # make relationships compatible with asyncio
    __mapper_args__ = {"eager_defaults": True}

    # relationships
    device_id = Column(Integer, ForeignKey('devices.id'))


    def __repr__(self):
        return "<Control ({})>".format(self.name)


    def as_dict(self):
        data = {
            'id': self.id,
            'name': self.name,
            'created': self.created.isoformat() + "Z",
            'device_id': self.device_id,
            'outlet': self.outlet,
            'attributes': self.attributes,
        }

        return data


    async def input(self, data):
        """
        Accept measurement data as input, evaluate, and perform on/off control
        as required:
        args:
            - dict (measurement data
        returns:
        """

        if self._precheck(data) is not True:
            return

        states = {'on': 1, 'off': 0}
        for key in ('on', 'off'):
            if (key not in self.attributes or not self.attributes[key] or
                self.attributes[key]['condition'] in [None, "", "null"] or
                self.attributes[key]['threshold'] in [None, "", "null"]):
                continue

            working = self.attributes[key]
            equation = "%0.1f %s %s" % (
                float(data['value']),
                working['condition'],
                working['threshold'])

            if evaluate(equation) is True:
                state = states[key]

                # switch device to state
                t1 = asyncio.create_task(self.switch_task(
                    self.device_id,
                    self.outlet,
                    state))

                # timed switch? set back to original state after seconds...
                if 'seconds' in working and int(working['seconds']) > 0:
                    t2 = asyncio.create_task(self.switch_task(
                        self.device_id,
                        self.outlet,
                        int(not state),
                        int(working['seconds'])
                    ))


    def _precheck(self, data):
        """
        examine incoming data, decide if it is valid for us
        """

        try:
            if data['device_id'] != self.attributes['input_device_id']:
                return False
        except Exception as x:
            logger.warning(str(x))
            return False

        try:
            if data['type'] != self.attributes['type']:
                return False
        except Exception as x:
            logger.warning(str(x))
            return False

        return True


    async def switch_task(self, device: int, outlet: int, state: int, delay: int = 0):
        try:
            if delay:
                await asyncio.sleep(delay)
                await switch_device_outlet(device, outlet, state)
            else:
                await switch_device_outlet(device, outlet, state)
        except Exception as x:
            logger.warning(x)



import re
import logging
import copy
from potnanny.plugins.base import (BluetoothDevicePlugin, GPIODevicePlugin,
    ActionPlugin, PipelinePlugin)
from potnanny.models.interface import ObjectInterface
from potnanny.models.keychain import Keychain
logger = logging.getLogger(__name__)


class InterfaceMixin:
    """
    Many Models have an "interface" attribute, which is a string reference to
    a plugin class. This mixin searches for the actual plugin class
    """

    @classmethod
    def interface_class(cls, name):
        klass = None
        parents = [BluetoothDevicePlugin, GPIODevicePlugin,
            ActionPlugin, PipelinePlugin]

        for p in parents:
            klass = p.get_named_class(name)
            if klass is not None:
                break

        return klass


class PluginMixin(InterfaceMixin):
    """
    Class for interfacing with Plugin object instance via the $parent.plugin property
    """

    @property
    def plugin(self):
        """
        Attribute to fetch an instantiated copy of our interfaces plugin.

        Required object attributes:
            - interface = like 'input.ble.govee.Hygrometer'
            - attributes = like {'address': '11:22:33:44:55:66'}

        Now, can communicate with the instance directly with initialized
        instance with self.plugin
        """

        if not hasattr(self, '_plugin'):
            self._plugin = None

        try:
            klass = self.interface_class(self.interface)
            self._plugin = klass(**self.attributes)
        except:
            pass

        try:
            # set up plugin specific introspection attributes, like:
            #   - is_bluetooth,
            #   - is_gpio,
            #   - is_reader,
            #   - is_pollable,
            #   - is_switchable,
            #   - is_key_required, etc

            if hasattr(self._plugin, 'poll') and callable(self._plugin.poll):
                setattr(self._plugin, 'is_pollable', True)

            if isinstance(self._plugin, BluetoothDevicePlugin):
                setattr(self._plugin, 'is_bluetooth', True)
                if (hasattr(self._plugin, 'read_advertisement') and
                    callable(self._plugin.read_advertisement)):
                    setattr(self._plugin, 'is_reader', True)
            elif isinstance(self._plugin, GPIODevicePlugin):
                setattr(self._plugin, 'is_gpio', True)

            if (hasattr(self._plugin, 'set_state') and
                callable(self._plugin.set_state)):
                setattr(self._plugin, 'is_switchable', True)

            if hasattr(self._plugin, 'key_code'):
                setattr(self._plugin, 'is_key_required', True)
        except:
            pass

        return self._plugin


class KeychainMixin:
    """
    A mixin class for a plugin (especially plugins that inherit from the
    ActionPlugin class) that must be associated with a named keychain.
    This requires the plugin has a self._keychain_name attribute.
    The keychain must be loaded async during setup:
        await self._load_my_keychain()

    Then, the plugin can access it's keychain data like:
        self.keychain.attributes
    """

    async def _load_my_keychain(self):
        if not hasattr(self, '_keychain'):
            self._keychain = None

        if not hasattr(self, '_keychain_name'):
            self._keychain_name = ''

        if self._keychain is None:
            self._keychain = await ObjectInterface(Keychain).get_by_name(
                self._keychain_name)

    @property
    def keychain(self):
        return self._keychain


class FingerprintMixin:
    """
    This is a class mixin for BLE plugins, to be able to recognize a
    BLE device that can be communicated with (by its device address or name)

    Each entry in a class fingerprint definition must match, in order for
    this to return True.
    """

    @classmethod
    def recognize_this(cls, fp):
        if not hasattr(cls, 'fingerprint'):
            logger.warning("Class %s no fingerprint defined" % cls)
            return False

        for key, regex in cls.fingerprint.items():
            if key not in fp:
                return False

            if not regex.search(fp[key]):
                return False

        return True



import os
import unittest
from potnanny.utils import evaluate

class TestUtilFuncs(unittest.TestCase):
# ###############################################

    def test_char_evals(self):
        assert evaluate("1 > 0") is True
        assert evaluate("0 < 1") is True
        assert evaluate("1 == 1") is True
        assert evaluate("1 >= 0") is True
        assert evaluate("0 <= 1") is True
        assert evaluate("0 != 1") is True
        assert evaluate("1 = 1") is True

    def test_str_evals(self):
        assert evaluate("1 gt 0") is True
        assert evaluate("0 lt 1") is True
        assert evaluate("1 eq 1") is True
        assert evaluate("1 ge 0") is True
        assert evaluate("0 le 1") is True
        assert evaluate("0 ne 1") is True


if __name__ == '__main__':
    unittest.main()
